import contextlib
import fcntl
import functools
import hashlib
import io
import json
import os
import random
import shutil
import site
import subprocess
import sys
import threading
import time
import traceback
from pathlib import Path
from typing import Optional, Dict, List, Callable

import mlflow
import pandas as pd
from dotenv import dotenv_values
from mlflow.models import infer_signature
from starlette.testclient import TestClient

from mlrpc.proto import base64_to_dir, ResponseObject, RequestObject, RequestObjectEncoded, KeyGenerator, \
    EncryptDecrypt, DataFileChunkWriter, Chunk

MLRPC_ENV_VARS_PRELOAD_KEY = "MLRPC_ENV_VARS_PRELOAD"


def request_to_df(request: RequestObject):
    return pd.DataFrame([request.encode().dict()])


def response_to_df(response: ResponseObject):
    return pd.DataFrame([response.encode().dict()])


def make_request_from_input_df(input_df: pd.DataFrame) -> List['RequestObject']:
    request_objs_enc = [RequestObjectEncoded(**row) for row in input_df.to_dict(orient='records')]
    return [RequestObject.from_request_enc(enc) for enc in request_objs_enc]


def load_mlrpc_env_vars() -> None:
    env_vars = os.getenv(MLRPC_ENV_VARS_PRELOAD_KEY, "")
    buff = io.StringIO(env_vars)
    loaded_env_vars = dotenv_values(stream=buff)
    for k, v in loaded_env_vars.items():
        os.environ[k] = v


def pack_env_file_into_preload(envfile: Path, env_dict: Dict[str, str] = None):
    env_dict = env_dict or os.environ
    if envfile.exists() is True:
        with envfile.open("r") as f:
            env_vars = f.read()
        env_dict[MLRPC_ENV_VARS_PRELOAD_KEY] = env_vars


def get_iso_datetime_timezone() -> str:
    return time.strftime("%Y-%m-%dT%H:%M:%S%z")


def debug_msg(msg: str, msg_type: str = "DEFAULT", level: str = "INFO"):
    ts = get_iso_datetime_timezone()
    level = f'[{level}]'
    msg_type = f'[{msg_type}]'.ljust(18)
    worker = f"[WID:{os.getpid()}]"
    print(f"[{ts}] {level} {msg_type} - {worker} {msg}", flush=True)


def copy_files(src: Path, dest: Path, check_dest_empty: bool = True):
    # copying due to not wanting requirements for pathspec
    src_dir = str(src)
    dest_dir = str(dest)

    if dest.exists() and any(dest.iterdir()) and check_dest_empty is True:
        debug_msg("Destination directory is not empty. Skipping file copy.", msg_type="COPY_FILES", level="INFO")
        return

    # clean the destination directory
    if dest.exists():
        shutil.rmtree(dest)

    dest.mkdir(parents=True, exist_ok=True)

    for root, dirs, files in os.walk(src_dir):
        for directory in dirs:
            dest_subdir = dest_dir / Path(root).relative_to(src_dir) / directory
            dest_subdir.mkdir(parents=True, exist_ok=True)

        for file in files:
            file_path = str(os.path.join(root, file))
            dest_file_path = str(dest_dir / Path(root).relative_to(src_dir) / file)
            debug_msg(f"Copying {file_path} to {dest_file_path}", msg_type="COPY_FILES", level="INFO")
            shutil.copy(file_path, dest_file_path)


class AppClientProxy:

    def __init__(self, app):
        self._app = app
        self._client = TestClient(app)

    @property
    def app(self):
        return self._app

    @property
    def client(self):
        return self._client

    def update_app(self, app):
        self._app = app
        self._client = TestClient(app)


class ReloadIndicator:

    def __init__(self, reload_indicator_location: str = None, reload_indicator_file: str = None):
        self._reload_indicator_file = Path(reload_indicator_file or "reload-indicator.txt")
        self._reload_indicator_dir = Path(reload_indicator_location or "/tmp/mlrpc-reload-indicator")
        self._reload_indicator = Path(self._reload_indicator_dir) / self._reload_indicator_file
        self._last_reload_info = None
        # make sure to boot up if the reload indicator exists
        if self.does_reload_indicator_exist() is True:
            self._last_reload_info = self.read_reload_indicator()

    def does_reload_indicator_exist(self) -> bool:
        return self._reload_indicator.exists()

    def _attempt_write(self, content: Optional[str] = None, attempts: int = 3) -> None:
        while attempts > 0:
            try:
                with self._reload_indicator.open("w") as f:
                    current_timestamp = get_iso_datetime_timezone()
                    f.write(content or current_timestamp)
                break
            except Exception as e:
                debug_msg(f"Failed to write to reload indicator: {e}", msg_type="RELOAD_INDICATOR", level="ERROR")
                attempts -= 1
                # random wait time to avoid multiple threads writing at the same time
                time.sleep(random.uniform(1, 3))

    def signal_reload(self, content: Optional[str] = None) -> None:
        debug_msg("Signaling reload", msg_type="RELOAD_INDICATOR", level="INFO")
        if self._reload_indicator_dir.exists() is False:
            self._reload_indicator_dir.mkdir(parents=True, exist_ok=True)

        self._attempt_write(content)

    def read_reload_indicator(self) -> Optional[str]:
        if self.does_reload_indicator_exist() is False:
            return None
        with self._reload_indicator.open("r") as f:
            return f.read()

    def should_i_reload(self) -> bool:
        if self._last_reload_info is None and self.does_reload_indicator_exist() is False:
            return False
        if self.does_reload_indicator_exist() is True and self._last_reload_info != self.read_reload_indicator():
            return True

        return False

    def update_last_reload_info(self):
        self._last_reload_info = self.read_reload_indicator()
        return self._last_reload_info


def make_reload_thread(reload_indicator: ReloadIndicator,
                       hot_reload_dispatcher: 'HotReloadEventHandler') -> threading.Thread:
    def reload_checker():
        debug_msg("Starting reload checker", msg_type="RELOAD_CHECKER", level="INFO")
        while True:
            if reload_indicator.should_i_reload() is True:
                debug_msg(f"Found reload indicator. Reloading app", msg_type="RELOAD_CHECKER",
                          level="INFO")
                try:
                    hot_reload_dispatcher.reload_app()
                except Exception as e:
                    debug_msg(f"Failed to reload app: {e}", msg_type="RELOAD_CHECKER", level="ERROR")
                finally:
                    reload_indicator.update_last_reload_info()
            time.sleep(5)

    return threading.Thread(target=reload_checker, daemon=True)


class HotReloadEventHandler:
    PATH_PREFIX = "__INTERNAL__"
    VALID_EVENTS = ["FULL_SYNC", "RELOAD", "REINSTALL", "RESET", "GET_PUBLIC_KEY", "UPLOAD_LARGE_FILE"]

    def __init__(self,
                 *,
                 app_client_proxy: AppClientProxy,
                 reload_code_path: str,
                 file_in_code_path: str,
                 obj_name: str,
                 reset_code_path: Optional[str] = None,
                 temp_key_dir: Optional[str] = "/tmp/mlrpc-hot-reload-keys",
                 dont_delete_data_dir: Optional[str] = None,
                 bootstrap_script_callback: Optional[Callable] = None
                 ):
        self._bootstrap_script_callback = bootstrap_script_callback
        self._reset_code_path = reset_code_path
        self._obj_name = obj_name
        self._file_in_code_path = file_in_code_path
        self._app_client_proxy = app_client_proxy
        self._reload_code_path = reload_code_path
        self._key_generator = KeyGenerator(temp_key_dir,
                                           debug_msg=functools.partial(debug_msg, msg_type="RSA_KEYGEN"))
        self._encrypt_decrypt = EncryptDecrypt(key_generator=self._key_generator)
        self._reload_indicator = ReloadIndicator()
        self._reload_thread = make_reload_thread(self._reload_indicator, self)
        self._reload_thread.start()
        self._data_chunk_writer = DataFileChunkWriter(
            root_dir=self._reload_code_path,
        )
        self._dont_delete_data_dir = dont_delete_data_dir

    def validate(self):
        debug_msg(
            f"Validating hot reload dispatcher with {self._app_client_proxy.app} {self._reload_code_path} "
            f"{self._file_in_code_path} {self._obj_name}",
            msg_type="HOT_RELOAD_VALIDATION",
            level="INFO")
        if not self._app_client_proxy.app:
            raise ValueError("app_proxy must be set")
        if not self._reload_code_path:
            raise ValueError("code_path must be set")
        if not self._file_in_code_path:
            raise ValueError("file_in_code_path must be set")
        if not self._obj_name:
            raise ValueError("obj_name must be set")

    def _is_valid_event(self, path: str) -> bool:
        normalized_req_path = path.lstrip("/").rstrip("/")
        valid_events = [f"{self.PATH_PREFIX}/{event}" for event in self.VALID_EVENTS]
        return normalized_req_path in valid_events

    def reload_app(self):
        debug_msg("Reloading bootstrap script!", msg_type="HOT_RELOAD", level="INFO")
        self._bootstrap_script_callback()
        import site
        import importlib
        site.addsitedir(self._reload_code_path)
        app_module = __import__(self._file_in_code_path.replace('.py', '').replace('/', '.'),
                                fromlist=[self._obj_name])
        importlib.reload(app_module)
        app_obj = getattr(app_module, self._obj_name)
        self._app_client_proxy.update_app(app_obj)
        debug_msg("Reloaded App!", msg_type="HOT_RELOAD", level="INFO")

    @staticmethod
    def _validate_checksum(content: str, checksum: str) -> bool:
        content_hash = hashlib.md5(content.encode('utf-8')).hexdigest()
        debug_msg("Valid checksum", msg_type="HOT_RELOAD", level="INFO")
        return content_hash == checksum

    def _do_full_sync(self, request: RequestObject):
        payload = json.loads(request.content)
        content = self._try_decrypt(payload['content'])
        checksum = payload['checksum']
        if self._validate_checksum(content, checksum) is False:
            return ResponseObject(
                status_code=400,
                content="Checksum validation failed"
            )
        skip_dirs = [self._dont_delete_data_dir] if self._dont_delete_data_dir else None
        base64_to_dir(content, self._reload_code_path,
                      skip_certain_directories=skip_dirs,
                      debug_msg=debug_msg)
        # send reload signal after files get moved
        self._reload_indicator.signal_reload()
        return ResponseObject(
            status_code=200,
            content="SUCCESS"
        )

    def _large_file_save(self, request: RequestObject):
        payload = json.loads(request.content)
        chunk = Chunk(**payload['chunk'])
        # TODO: get checksum from payload in future
        if chunk.start_of_file:
            debug_msg(f"Starting new file: {chunk.relative_file_path}", msg_type="UPLOAD_LARGE_FILE", level="INFO")
        if chunk.eof:
            debug_msg(f"Ending file: {chunk.relative_file_path}", msg_type="UPLOAD_LARGE_FILE", level="INFO")
        self._data_chunk_writer.write_chunk(chunk)
        return ResponseObject(
            status_code=200,
            content="SUCCESS"
        )

    @staticmethod
    def _install_packages(package_names: List[str]):
        try:
            subprocess.check_call([sys.executable, "-m", "pip", "install", *package_names])
            return f"Packages {','.join(package_names)} installed successfully"
        except subprocess.CalledProcessError:
            raise ValueError(f"Failed to install packages {','.join(package_names)}")

    @staticmethod
    def _uninstall_package(package_name: str):
        try:
            subprocess.check_call([sys.executable, "-m", "pip", "uninstall", "-y", package_name])
            return f"Package {package_name} uninstalled successfully"
        except subprocess.CalledProcessError:
            f"Failed to uninstall package {package_name}"

    def _do_install(self, request: RequestObject):
        payload = json.loads(request.content)
        requirements = payload['requirements']
        install_message = self._install_packages(requirements)
        # send reload signal after python libs are installed
        self._reload_indicator.signal_reload()
        return ResponseObject(
            status_code=200,
            content=json.dumps(install_message)
        )

    def _try_decrypt(self, something: Optional[str]) -> Optional[str]:
        if something is None:
            return None
        if isinstance(something, str) is True:
            try:
                return self._encrypt_decrypt.decrypt(something)
            except Exception as e:
                debug_msg(f"Failed to decrypt: {str(e)}", msg_type="CONTENT_DECRYPTION", level="ERROR")
                return something
        return something

    def dispatch(self, request: RequestObject) -> Optional[ResponseObject]:
        if request.method != "POST":
            return None

        if self._is_valid_event(request.path) is False:
            return None

        if request.path == f"/{self.PATH_PREFIX}/FULL_SYNC":
            debug_msg("Dispatching full sync event", msg_type="FULL_SYNC", level="INFO")
            self._do_full_sync(request)
            return ResponseObject(
                status_code=200,
                content="SUCCESS"
            )

        if request.path == f"/{self.PATH_PREFIX}/REINSTALL":
            debug_msg("Dispatching reinstall event", msg_type="REINSTALL_REQS", level="INFO")
            return self._do_install(request)

        if request.path == f"/{self.PATH_PREFIX}/GET_PUBLIC_KEY":
            debug_msg("Dispatching get public key event", msg_type="GET_PUBLIC_KEY", level="INFO")
            return ResponseObject(
                status_code=200,
                content=json.dumps({
                    "public_key": self._key_generator.get_public_key()
                })
            )

        if request.path == f"/{self.PATH_PREFIX}/UPLOAD_LARGE_FILE":
            debug_msg("Dispatching large file upload event", msg_type="UPLOAD_LARGE_FILE", level="INFO")
            return self._large_file_save(request)


@contextlib.contextmanager
def app_directory(destination):
    prev_dir = os.getcwd()
    os.chdir(destination)
    try:
        yield
    finally:
        os.chdir(prev_dir)


def try_become_leader(
        lock_file_path: Path,
        pid_file_path: Path,
        bootstrap_script_path: Path,
        sleep_waiting_for_leader: int = 5
):
    try:
        with open(lock_file_path, "w") as lock_file:
            fcntl.flock(lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
            time.sleep(sleep_waiting_for_leader)  # retain the lock for 5 seconds
            debug_msg("I was elected the leader. Running the bootstrap script...",
                      msg_type="BOOTSTRAP_SCRIPT", level="INFO")
            debug_msg("Releasing the leader lock...",
                      msg_type="BOOTSTRAP_SCRIPT", level="INFO")
            run_bootstrap_script(
                pid_file_path,
                bootstrap_script_path
            )
            # unlock after bootstrap is initialized
            fcntl.flock(lock_file, fcntl.LOCK_UN)
        try:
            lock_file_path.unlink()
        except FileNotFoundError:
            debug_msg("Lock file was already deleted by another process.",
                      msg_type="BOOTSTRAP_SCRIPT", level="INFO")

    except BlockingIOError:
        debug_msg("Not the leader, proceeding as a regular worker.",
                  msg_type="BOOTSTRAP_SCRIPT", level="INFO")


def cleanup_if_exists(pid_file_path: Path):
    if pid_file_path.exists():
        with open(pid_file_path, 'r') as f:
            pid = int(f.read())
            debug_msg(f"Killing the bootstrap script with PID: {pid}",
                      msg_type="BOOTSTRAP_SCRIPT", level="INFO")
            subprocess.run(f"kill -9 {pid}", shell=True)
            pid_file_path.unlink()
        debug_msg("Cleaned up the pid file.",
                  msg_type="BOOTSTRAP_SCRIPT", level="INFO")


def run_bootstrap_script(pid_file_path: Path, bootstrap_script_path: Path):
    cleanup_if_exists(pid_file_path)
    debug_msg("Running the bootstrap script...",
              msg_type="BOOTSTRAP_SCRIPT", level="INFO")
    process = subprocess.Popen([sys.executable, str(bootstrap_script_path)],
                               start_new_session=True)
    pid = process.pid
    with open(pid_file_path, 'w') as f:
        f.write(str(pid))
    debug_msg(f"Bootstrap script is running with PID: {pid}",
              msg_type="BOOTSTRAP_SCRIPT", level="INFO")


class FastAPIFlavor(mlflow.pyfunc.PythonModel):

    def __init__(self,
                 *,
                 local_app_dir_abs: Optional[str],
                 local_app_path_in_dir: Optional[str] = "app.py",
                 app_obj: Optional[str] = "app",
                 artifact_code_key="code",
                 reloadable: Optional[bool] = False,
                 data_dir: Optional[str] = None,
                 bootstrap_script: Optional[str] = None):
        self._reloadable = reloadable
        self.local_app_dir = local_app_dir_abs
        self.code_key = artifact_code_key
        self.local_app_dir = self.local_app_dir.rstrip("/")
        self.app_path_in_dir = local_app_path_in_dir
        self.app_obj = app_obj
        self._app_proxy = None
        self._hot_reload_dispatcher: Optional[HotReloadEventHandler] = None
        self._app_working_directory = None
        self._dont_delete_data_dir = data_dir
        self._bootstrap_script = bootstrap_script

    def load_module(self, app_dir_mlflow_artifacts: Optional[str] = None):
        with app_directory(self._app_working_directory):
            app_dir = app_dir_mlflow_artifacts or self.local_app_dir
            # only add if it's not already there
            site.addsitedir(app_dir)
            debug_msg(f"Loading code from path {app_dir}", msg_type="LOAD_MODULE", level="INFO")
            app_module = __import__(self.app_path_in_dir.replace('.py', '').replace('/', '.'), fromlist=[self.app_obj])
            debug_msg(f"Loaded module successfully for {self.app_path_in_dir} from {app_dir}",
                      msg_type="LOAD_MODULE", level="INFO")
            app_obj = getattr(app_module, self.app_obj)
            debug_msg(f"Loaded app object successfully for {self.app_obj}",
                      msg_type="LOAD_MODULE", level="INFO")
            return app_obj

    # def validate(self):
    #     if not self.local_app_dir:
    #         raise ValueError("local_app_dir_abs must be set")
    #     if not self.app_path_in_dir:
    #         raise ValueError("app_path_in_dir must be set")
    #     if not self.app_obj:
    #         raise ValueError("app_obj must be set")
    #     try:
    #         self.load_module()
    #     except Exception as e:
    #         raise ValueError(f"Failed to load app module: {e} in {self.local_app_dir}/{self.app_path_in_dir}")

    def _run_bootstrap_script(self, code_dir: Path | str,
                              delay: int = 5
                              ):
        if isinstance(code_dir, str):
            code_dir = Path(code_dir)

        bootstrap_stage_dir = "/tmp/mlrpc-bootstrap"
        bootstrap_stage_filename = "bootstrap.py"
        if self._bootstrap_script is None:
            return
        bootstrap_script_path = Path(code_dir) / self._bootstrap_script
        if not bootstrap_script_path.exists():
            return

        # stage bootstrap script
        Path(bootstrap_stage_dir).mkdir(parents=True, exist_ok=True)
        bootstrap_staged_file = Path(bootstrap_stage_dir, bootstrap_stage_filename)
        shutil.copy(bootstrap_script_path, bootstrap_staged_file)

        with app_directory(bootstrap_stage_dir):
            lock_file_path = Path(bootstrap_stage_dir, "leader.lock")
            pid_file_path = Path(bootstrap_stage_dir, "bootstrap.pid")
            try_become_leader(lock_file_path,
                              pid_file_path,
                              bootstrap_staged_file,
                              # we want the leader to wait 5 seconds roughly
                              # before unlocking the leader election file
                              # TODO: remove this in future
                              sleep_waiting_for_leader=delay
                              )

    def load_context(self, context):
        debug_msg("Loading preloaded env vars",
                  msg_type="LOAD_CONTEXT", level="INFO")
        load_mlrpc_env_vars()
        code_path = self.local_app_dir
        if context is not None:
            code_path = context.artifacts[self.code_key]

        self._app_proxy = AppClientProxy(None)
        self._app_working_directory = code_path

        # add the app_path to the pythonpath to load modules
        # only if it's not already there
        # update site packages to include this app dir

        def boot_app(_temp_code_dir, _temp_key_dir):
            Path(_temp_code_dir).mkdir(parents=True, exist_ok=True)
            Path(_temp_key_dir).mkdir(parents=True, exist_ok=True)
            debug_msg("Hot reload dir being created at /tmp/mlrpc-hot-reload",
                      msg_type="HOT_RELOAD_SETUP", level="INFO")

            copy_files(Path(code_path),
                       Path(_temp_code_dir),
                       check_dest_empty=True, )
            debug_msg(f"Copying files from {code_path} to {_temp_code_dir}",
                      msg_type="HOT_RELOAD_SETUP", level="INFO")
            self._hot_reload_dispatcher = HotReloadEventHandler(
                app_client_proxy=self._app_proxy,
                reload_code_path=_temp_code_dir,
                file_in_code_path=self.app_path_in_dir,
                obj_name=self.app_obj,
                reset_code_path=code_path,
                temp_key_dir=_temp_key_dir,
                dont_delete_data_dir=self._dont_delete_data_dir,
                bootstrap_script_callback=functools.partial(self._run_bootstrap_script, _temp_code_dir, delay=0)
            )
            debug_msg(f"Hot reload dispatcher created for {self.app_obj} in {self.app_path_in_dir}",
                      msg_type="HOT_RELOAD_SETUP", level="INFO")
            self._app_proxy.update_app(self.load_module(app_dir_mlflow_artifacts=_temp_code_dir))

        if os.getenv("MLRPC_HOT_RELOAD", str(self._reloadable)).lower() == "true":
            temp_code_dir = "/tmp/mlrpc-hot-reload"
            temp_key_dir = "/tmp/mlrpc-hot-reload-keys"
            self._app_working_directory = temp_code_dir
            attempts = 5
            while True:
                try:
                    boot_app(temp_code_dir, temp_key_dir)
                    break
                except Exception as e:
                    debug_msg(f"Error occurred while booting app: {e}",
                              msg_type="LOAD_CONTEXT", level="ERROR")
                    attempts -= 1
                    wait_time = random.uniform(1, 5)
                    time.sleep(wait_time)
                    if attempts == 0:
                        raise
                    debug_msg(f"Retrying in 5 seconds",
                              msg_type="LOAD_CONTEXT", level="INFO")
            # all code is copied and temp_code_dir is source of truth
            self._run_bootstrap_script(temp_code_dir)
        else:
            self._run_bootstrap_script(code_path)
            self._app_proxy.update_app(self.load_module(app_dir_mlflow_artifacts=code_path))

    def predict(self, context, model_input: pd.DataFrame, params=None):
        if self._app_proxy is None:
            self._app_proxy = AppClientProxy(None)

        if self._app_proxy.app is None or self._app_proxy.client is None:
            self.load_context(context)

        requests = make_request_from_input_df(model_input)
        responses = []

        try:
            # happy path things can go wrong :-)
            with app_directory(self._app_working_directory):
                if self._hot_reload_dispatcher is not None and len(requests) == 1:
                    event_resp = self._hot_reload_dispatcher.dispatch(requests[0])
                    if event_resp is not None:
                        return response_to_df(event_resp)

                for req in requests:
                    resp = self._app_proxy.client.request(
                        method=req.method,
                        url=req.path,
                        headers=req.headers,
                        params=req.query_params,
                        data=req.content,
                        timeout=req.timeout
                    )
                    responses.append(ResponseObject.from_httpx_resp(resp).encode())

            return pd.DataFrame([resp.dict() for resp in responses])
        except Exception as e:
            traceback.print_exc()
            debug_msg(f"Error occurred: {e}", msg_type="PREDICT", level="ERROR")
            return response_to_df(ResponseObject(
                status_code=500,
                content=f"Error occurred during predict look at server logs."
            ))

    @staticmethod
    def signature():
        full_request = request_to_df(RequestObject(
            method='GET',
            headers=[('Host', 'example.org')],
            path='/',
            query_params="",
            content="",
            timeout=10
        ))
        optional_request = request_to_df(RequestObject(
            method='GET',
            path='/'
        ))
        full_response = response_to_df(ResponseObject(
            status_code=200,
            headers=[('Content-Type', 'application/json')],
            content=""
        ))
        optional_response = response_to_df(ResponseObject(
            status_code=200,
        ))
        return infer_signature(
            model_input=pd.concat([full_request, optional_request]),
            model_output=pd.concat([full_response, optional_response])
        )



import os
import shutil
from datetime import timedelta
from importlib.metadata import distributions
from pathlib import Path
from typing import Optional, List, Literal

import mlflow
from databricks.sdk import WorkspaceClient
from databricks.sdk.service.catalog import ModelVersionInfo
from databricks.sdk.service.serving import EndpointCoreConfigInput, ServedEntityInput, EndpointStateReady, \
    TrafficConfig, Route
from mlflow.entities import Experiment
from mlflow.entities.model_registry import ModelVersion
from pathspec import PathSpec
from pathspec.patterns import GitWildMatchPattern

from mlrpc.flavor import FastAPIFlavor, MLRPC_ENV_VARS_PRELOAD_KEY
from mlrpc.utils import get_version


def get_home_directory(ws_client: WorkspaceClient) -> Path:
    user = ws_client.current_user.me().user_name
    return Path("/Users") / str(user)


def get_or_create_mlflow_experiment(ws_client: WorkspaceClient, experiment_name: str) -> Experiment:
    experiment_path = get_home_directory(ws_client) / experiment_name
    experiment_path_str = str(experiment_path)
    exp = mlflow.get_experiment_by_name(experiment_path_str)
    if exp is None:
        print("Creating experiment")
        exp = mlflow.create_experiment(experiment_path_str)
        return mlflow.get_experiment(exp)
    return exp


def ensure_3_parts(name: str) -> bool:
    return len(name.split(".")) == 3


class LibraryManager:

    def __init__(self):
        self._installed_libraries = {
            i.metadata['name']: i.version for i in distributions()
        }

    def has_library(self, library_name) -> bool:
        return library_name in self._installed_libraries

    def get_library_version(self, library_name):
        return self._installed_libraries.get(library_name)

    def library_pinned_string(self, library_name):
        return f"{library_name}=={self.get_library_version(library_name)}"


def default_mlrpc_libs():
    lbm = LibraryManager()
    libs = []
    important_libs = ["cloudpickle",
                      "mlflow-skinny",
                      "fastapi",
                      "pandas",
                      "httpx",
                      "python-dotenv",
                      "cryptography"]
    mlrpc_req = "mlrpc"
    mlflow_req = "mlflow"
    if lbm.has_library(mlrpc_req) is False or len(lbm.get_library_version(mlrpc_req).split(".")) > 3:
        libs.append("mlrpc")
    else:
        libs.append(lbm.library_pinned_string(mlrpc_req))

    if lbm.has_library(mlflow_req) is False:
        if lbm.has_library("mlflow-skinny") is True:
            libs.append(lbm.library_pinned_string("mlflow-skinny"))
            libs.append(lbm.library_pinned_string("mlflow-skinny").replace("mlflow-skinny", "mlflow"))
        else:
            libs.append("mlflow")
    else:
        libs.append(lbm.library_pinned_string(mlflow_req))

    for lib in important_libs:
        if lbm.has_library(lib) is False:
            libs.append(lib)
        else:
            libs.append(lbm.library_pinned_string(lib))

    return libs


def ensure_mlflow_installation(pip_reqs: List[str]):
    if any(["mlflow==" not in pip_req for pip_req in pip_reqs]):
        mlflow_pip_install = None
        for pip_req in pip_reqs:
            if pip_req.startswith("mlflow-skinny=="):
                mlflow_pip_install = pip_req.replace("mlflow-skinny", "mlflow")
                break
        if mlflow_pip_install is None:
            raise ValueError("mlflow-skinny==<version> must be present in pip_requirements")
        pip_reqs.append(mlflow_pip_install)
    return pip_reqs


def build_proper_requirements(file_path: Path) -> List[str]:
    req_libs = ensure_mlflow_installation(default_mlrpc_libs())
    if file_path.exists() is False:
        return req_libs

    print("Found requirements.txt using this")
    from pkg_resources import parse_requirements
    prebuilt_reqs = [lib.split("==")[0] for lib in req_libs]

    with open(file_path, 'r') as file:
        lines = file.readlines()
        # Ignore comments and empty lines
        lines = [line.strip() for line in lines if not line.startswith('#') and line.strip() != '']
        requirements = [str(req) for req in parse_requirements(lines) if req.name not in prebuilt_reqs]
        return req_libs + requirements


def stage_files_for_deployment(app: FastAPIFlavor, dest_path: Optional[str] = None):
    src = Path(app.local_app_dir)
    dest = Path(dest_path)
    ignore_file = src / ".gitignore"
    ignore_file = ignore_file if ignore_file.exists() else None
    copy_files(app.local_app_dir, dest, ignore_file)


def save_model(
        ws_client: WorkspaceClient,
        experiment: Experiment,
        app: FastAPIFlavor,
        uc_model_path: str,
        run_name: str = "deployment",
        code_path: Optional[str] = None,
        aliases: Optional[List[str]] = None,
        latest_alias_name: Optional[str] = "current",
        reload: bool = False,
) -> ModelVersion:
    if not ensure_3_parts(uc_model_path):
        raise ValueError \
            (f"Model path must be in the format 'catalog_name.schema_name.model_name' but got {uc_model_path}")

    aliases = aliases or []
    aliases.append(latest_alias_name)
    code_path = Path(code_path)
    potential_requirements_file = code_path / "requirements.txt"
    requirements = build_proper_requirements(potential_requirements_file)

    with mlflow.start_run(experiment_id=experiment.experiment_id, run_name=run_name) as run:
        mlflow.pyfunc.log_model(
            artifact_path="model",
            python_model=app,
            signature=app.signature(),
            artifacts={app.code_key: str(code_path)},
            pip_requirements=requirements
        )
        mlflow.set_tag("mlrpc_version", get_version("mlrpc"))
        mlflow.set_tag("reloadable", 'true' if reload else 'false')

    mv = mlflow.register_model(f"runs:/{run.info.run_id}/model",
                               uc_model_path,
                               tags={
                                   "mlrpc_version": get_version("mlrpc"),
                                   "reloadable": 'true' if reload else 'false'
                               })
    for alias in aliases:
        ws_client.registered_models.set_alias(
            full_name=uc_model_path,
            version_num=mv.version,
            alias=alias
        )
    return mv


def keep_only_last_n_versions(ws_client: WorkspaceClient, uc_model_path: str, n: int):
    versions = ws_client.model_versions.list(uc_model_path)
    versions = sorted(versions, key=lambda x: x.version, reverse=True)
    # delete any versions other than the last n versions
    if len(versions) > n:
        for version in versions[n:]:
            print("Deleting version", version.version)
            ws_client.model_versions.delete(uc_model_path, version.version)


def copy_files(src, dest, ignore_file=None):
    spec = None
    src_dir = str(src)
    dest_dir = str(dest)
    if ignore_file is not None:
        ignore_file = str(ignore_file)
        with open(ignore_file, 'r') as f:
            gitignore = f.read()
        spec = PathSpec.from_lines(GitWildMatchPattern, gitignore.splitlines())

    # clean the destination directory
    if dest.exists():
        shutil.rmtree(dest)

    for root, dirs, files in os.walk(src_dir):
        for directory in dirs:
            dest_subdir = dest_dir / Path(root).relative_to(src_dir) / directory
            dest_subdir.mkdir(parents=True, exist_ok=True)

        for file in files:
            file_path = str(os.path.join(root, file))
            dest_file_path = str(dest_dir / Path(root).relative_to(src_dir) / file)
            if ignore_file is None:
                print("Copying", file_path, "to", dest_file_path)
                Path(dest_file_path).parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(file_path, dest_file_path)
                continue

            if not spec.match_file(file_path):
                print("Copying", file_path, "to", dest_file_path)
                shutil.copy(file_path, dest_file_path)


def deploy_secret_env_file(ws_client: WorkspaceClient, secret_scope: str, secret_key: str, env_file: Path):
    if env_file.exists() is False:
        return
    with open(env_file, 'r') as file:
        string_value = file.read()
        ws_client.secrets.put_secret(secret_scope, secret_key, string_value=string_value)


def _check_deployable(ws_client: WorkspaceClient, endpoint_name: str) -> Literal["DEPLOY", "UPDATE", "NOT_UPDATABLE"]:
    try:
        endpoint = ws_client.serving_endpoints.get(endpoint_name)
        # return endpoint.state.ready == EndpointStateReady.READY
        if endpoint.state.ready == EndpointStateReady.READY:
            return "UPDATE"
        return "NOT_UPDATABLE"
    except Exception:
        return "DEPLOY"


def deploy_serving_endpoint(ws_client: WorkspaceClient,
                            endpoint_name: str,
                            uc_model_path: str,
                            model_version: ModelVersion | ModelVersionInfo,
                            reload_version: ModelVersion | ModelVersionInfo | None = None,
                            secret_scope: Optional[str] = None,
                            secret_key: Optional[str] = None,
                            workload_type: str = "CPU",
                            size: str = "Small",
                            scale_to_zero_enabled: bool = True):
    if _check_deployable(ws_client, endpoint_name) == "NOT_UPDATABLE":
        raise ValueError(f"Endpoint {endpoint_name} is not ready state to be updated")

    env_vars = None
    if secret_scope is not None and secret_key is not None:
        env_vars = {
            MLRPC_ENV_VARS_PRELOAD_KEY: "{{" + f"secrets/{secret_scope}/{secret_key}" + "}}"
        }

    reload_entities = []
    reload_traffic = []
    if reload_version is not None:
        reload_entities.append(ServedEntityInput(
            name="hotreload",
            entity_version=reload_version.version,
            entity_name=uc_model_path,
            workload_type=workload_type,
            workload_size="Small",  # workload size for hotreload should always be small
            scale_to_zero_enabled=scale_to_zero_enabled,
            environment_vars=env_vars
        ))
        reload_traffic.append(Route(
            served_model_name="hotreload",
            traffic_percentage=0,
        ))

    if _check_deployable(ws_client, endpoint_name) == "UPDATE":
        return ws_client.serving_endpoints.update_config_and_wait(
            name=endpoint_name,
            served_entities=[
                ServedEntityInput(
                    name="main",
                    entity_version=model_version.version,
                    entity_name=uc_model_path,
                    workload_size=size,
                    workload_type=workload_type,
                    scale_to_zero_enabled=scale_to_zero_enabled,
                    environment_vars=env_vars
                ),
                *reload_entities,
            ],
            traffic_config=TrafficConfig(
                routes=[
                    Route(
                        served_model_name="main",
                        traffic_percentage=100,
                    ),
                    *reload_traffic
                ]
            ),
            timeout=timedelta(minutes=60)
        )

    return ws_client.serving_endpoints.create_and_wait(
        name=endpoint_name,
        config=EndpointCoreConfigInput(
            name=endpoint_name,
            served_entities=[
                ServedEntityInput(
                    name="main",
                    entity_version=model_version.version,
                    entity_name=uc_model_path,
                    workload_size=size,
                    workload_type=workload_type,
                    scale_to_zero_enabled=scale_to_zero_enabled,
                    environment_vars=env_vars
                ),
                *reload_entities
            ],
            traffic_config=TrafficConfig(
                routes=[
                    Route(
                        served_model_name="main",
                        traffic_percentage=100,
                    ),
                    *reload_traffic
                ]
            ),
        ),
        timeout=timedelta(minutes=60)
    )



import os

from fastapi import FastAPI, Request
from fastapi.openapi.docs import get_swagger_ui_html
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import Response, JSONResponse


def make_swagger_proxy(
        endpoint_name: str,
        profile: str = "default",
        port: int = 8000,
        debug: bool = False,
        databricks_mode: bool = True,
        local_server_port: int = 6500,
        hotreload_enabled: bool = False,
):
    class RawRequestMiddleware(BaseHTTPMiddleware):
        async def dispatch(self, request: Request, call_next):
            # Load the raw request
            raw_request = await request.body()

            from mlrpc.client import rpc

            def get_just_path(request: Request):
                return str(request.url).replace(str(request.base_url), "/")

            path = get_just_path(request)
            if path.startswith("/docs"):
                return get_swagger_ui_html(
                    openapi_url=f"http://0.0.0.0:{port}/openapi.json",
                    title="MLRPC Swagger UI",

                )

            if databricks_mode is True:
                os.environ["DATABRICKS_CONFIG_PROFILE"] = profile
                rpc_flavor = rpc.databricks(endpoint_name)
                mlrpc_response = rpc_flavor \
                    .request(str(request.method), path, headers=request.headers, data=raw_request.decode("utf-8"))
            else:
                rpc_flavor = rpc.local(port=local_server_port)
                mlrpc_response = rpc_flavor \
                    .request(str(request.method), path, headers=[(k, v) for k, v in request.headers.items()],
                             data=raw_request.decode("utf-8"))
            headers = mlrpc_response.headers
            if headers is not None:
                headers = {k: v for d in mlrpc_response.headers for k, v in d.items() if k.lower() != "content-length"}
            if debug is True:
                print("MLRPC RESPONSE: ", mlrpc_response)
            if isinstance(mlrpc_response.body, str):
                content = mlrpc_response.body
                response = Response(content=content, status_code=mlrpc_response.status_code,
                                    headers=headers)
            else:
                if path.endswith("openapi.json") and hotreload_enabled is True:
                    if isinstance(mlrpc_response.body, dict):
                        info = mlrpc_response.body.get("info", {})
                        if info is not None:
                            info["title"] = info.get("title", "MLRPC Swagger UI").strip() + " - HOTRELOAD"
                        return JSONResponse(content=mlrpc_response.body, status_code=mlrpc_response.status_code,
                                                headers=headers)
                response = JSONResponse(content=mlrpc_response.body, status_code=mlrpc_response.status_code,
                                        headers=headers)

            return response

    app = FastAPI()

    @app.middleware("http")
    async def add_raw_request_middleware(request: Request, call_next):
        middleware = RawRequestMiddleware(app)
        return await middleware.dispatch(request, call_next)

    return app



import os
import re
from pathlib import Path
from typing import Generator

from detect_secrets import SecretsCollection
from detect_secrets.plugins.base import RegexBasedDetector
from detect_secrets.settings import transient_settings

path = os.path.realpath(__file__)


class DAPIDetector(RegexBasedDetector):
    """Scans for databricks api tokens."""
    secret_type = 'Databricks API Token'

    denylist = (
        re.compile(
            r'dapi[\w]*',
            re.IGNORECASE,
        ),
        re.compile(
            r'dkea[\w]*',
            re.IGNORECASE,
        ),
    )

    def analyze_string(self, string: str) -> Generator[str, None, None]:
        for regex in self.denylist:
            for match in regex.findall(string):
                if len(match) < 10:  # ignore short matches
                    continue
                if isinstance(match, tuple):
                    for submatch in filter(bool, match):
                        yield submatch
                else:
                    yield match


all_plugins = """ArtifactoryDetector
AWSKeyDetector
AzureStorageKeyDetector
BasicAuthDetector
CloudantDetector
DiscordBotTokenDetector
GitHubTokenDetector
Base64HighEntropyString
HexHighEntropyString
IbmCloudIamDetector
IbmCosHmacDetector
JwtTokenDetector
MailchimpDetector
NpmDetector
PrivateKeyDetector
SendGridDetector
SlackDetector
SoftlayerDetector
SquareOAuthDetector
StripeDetector
TwilioKeyDetector""".split("\n")


def get_all_files(directory):
    return [str(file) for file in Path(directory).rglob('*') if file.is_file()]


def scan_in_directory(directory: str) -> Generator[str, None, None]:
    files = get_all_files(directory)
    secrets = SecretsCollection()
    with transient_settings({
        # Only run scans with only these plugins.
        # This format is the same as the one that is saved in the generated baseline.
        'plugins_used': [
            {
                'name': 'DAPIDetector',
                'path': f"file://{path}"
            },
            *[{"name": p_name} for p_name in all_plugins]
        ],

    }):
        secrets.scan_files(
            *files, num_processors=None
        )
        for k, l in secrets.json().items():
            for s in l:
                relative_file = Path(s['filename']).relative_to(directory)
                yield f"POTENTIAL SECRET FOUND - [{s['type']}][LINENO:{s['line_number']}]: {relative_file}"



import os

import click
from click_configfile import ConfigFileReader, Param, SectionSchema
from click_configfile import matches_section


class ConfigSectionSchema(object):
    """Describes all config sections of this configuration file."""

    @matches_section("app")
    class Root(SectionSchema):
        name = Param(type=str)
        app_root_dir = Param(type=click.Path(), default=os.getcwd())
        app_path_in_root = Param(type=str, default="app.py")
        app_obj = Param(type=str, default="app")
        data_dir = Param(type=click.Path(exists=True, dir_okay=True), default=None)
        uc_catalog = Param(type=str)
        uc_schema = Param(type=str)
        endpoint_name = Param(type=str)
        latest_alias_name = Param(type=str, default="current")
        env = Param(type=str, default="dev")
        make_experiment = Param(type=bool, default=True)
        experiment_name = Param(type=str)
        register_model = Param(type=bool, default=True)
        databricks_profile = Param(type=str, default="default")
        only_last_n_versions = Param(type=int, default=None)
        secret_scope = Param(type=str, default=None)
        secret_key = Param(type=str, default=None)
        env_file = Param(type=click.Path(), default=None)
        type = Param(type=str, default="CPU")
        size = Param(type=str, default="Small")
        scale_to_zero_enabled = Param(type=bool, default=True)
        bootstrap_python_script = Param(type=click.Path(), default=None)


class ConfigFileProcessor(ConfigFileReader):
    config_files = ["mlrpc.ini", "mlrpc.cfg"] if os.getenv("MLRPC_CONFIG") is None else [os.getenv("MLRPC_CONFIG")]
    config_section_schemas = [
        ConfigSectionSchema.Root,  # PRIMARY SCHEMA
    ]


INIT_CONFIG = """## The following lines are minimum required config
## You need to have the app section
# [app]
# name = <app name>
# uc_catalog = <catalog name>
# uc_schema = <schema name>
# endpoint_name = <endpoint name>

# this is optional if you want to upload larger binaries like chroma, sqlite, faiss, lancedb, etc
# data_dir=data

# Cost controls
# Small, Medium, Large are valid values
# size = Small
# various values 
# refer to https://docs.databricks.com/en/machine-learning/model-serving/create-manage-serving-endpoints.html
# or https://learn.microsoft.com/en-us/azure/databricks/machine-learning/model-serving/create-manage-serving-endpoints
# type = CPU
# scale_to_zero_enabled = true # true, false

# This is an optional field if you want to run a python script before the main app
# this is great if you want to spawn a vllm process, deepspeed-mi server, tei, tgi, some app in another lang.
# bootstrap_python_script = <path to your python script>

## The following lines are optional and not required

# latest_alias_name = <alternative alias> uses "current" by default

## optional specify your experiment name
# experiment_name = <experiment name> 

## Define where your code is
# app_root_dir = <root of your directory>

## The following lines are optional if you need secrets
# secret_scope = <secret scope in databricks>
# secret_key = <secret key>
# env_file = <location of your env file relative to this file>

## use this setting if you want to delete old versions
# only_last_n_versions=10
"""



import base64
import io
import os
import socket
import subprocess
import sys
import tarfile
from dataclasses import dataclass
from pathlib import Path
from typing import List

from pathspec import PathSpec
from pathspec.patterns import GitWildMatchPattern


def ensure_python_path(env):
    for python_version_dir in (Path(sys.executable).parent.parent / "lib").iterdir():
        site_packages = str(python_version_dir / "site-packages")
        py_path = env.get("PYTHONPATH", "")
        if site_packages not in py_path.split(":"):
            env["PYTHONPATH"] = f"{py_path}:{site_packages}"


def execute(*, cmd: List[str], env, cwd=None, ensure_python_site_packages=True, shell=False, trim_new_line=True):
    if ensure_python_site_packages:
        ensure_python_path(env)
    if shell is True:
        cmd = " ".join(cmd)
    popen = subprocess.Popen(cmd,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             universal_newlines=True,
                             shell=shell,
                             env=env,
                             cwd=cwd,
                             bufsize=1)
    if popen.stdout is not None:
        for stdout_line in iter(popen.stdout.readline, ""):
            if trim_new_line:
                stdout_line = stdout_line.strip()
            yield stdout_line

    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def find_next_open_port(start_port: int = 6500, end_port: int = 7000) -> int:
    for port in range(start_port, end_port + 1):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            try:
                sock.bind(("0.0.0.0", port))
                return port
            except OSError:
                continue
    raise Exception(f"No open ports found in the range {start_port}-{end_port}")


@dataclass
class DatabricksProfile:
    name: str
    host: str
    token: str


def get_profile_contents(profile_name, profile_path: str = "~/.databrickscfg") -> DatabricksProfile:
    from configparser import ConfigParser
    import os
    profile_path = os.path.expanduser(profile_path)
    config = ConfigParser()
    config.read(profile_path)
    host = config[profile_name]["host"]
    token = config[profile_name]["token"]
    return DatabricksProfile(name=profile_name, host=host, token=token)


def get_version(package_name: str = "mlrpc") -> str:
    try:
        from importlib import metadata  # type: ignore
    except ImportError:
        # Python < 3.8
        import importlib_metadata as metadata  # type: ignore

    try:
        return metadata.version(package_name)  # type: ignore
    except metadata.PackageNotFoundError:  # type: ignore
        return "unknown"


def file_size_to_mb(file_size: int, _round: int) -> str:
    return f"{round(file_size / (1024 * 1024), _round)} MB"


def dir_to_base64(tar_dir, ignore_file: Path = None):
    # Create a BytesIO object
    data = io.BytesIO()

    spec = None
    if ignore_file is not None and ignore_file.exists() and ignore_file.is_file():
        ignore_file = str(ignore_file)
        with open(ignore_file, 'r') as f:
            gitignore = f.read()
        spec = PathSpec.from_lines(GitWildMatchPattern, gitignore.splitlines())

    total_size = 0  # total size of the files
    with tarfile.open(fileobj=data, mode='w:gz') as tar:
        for root, dirs, files in os.walk(tar_dir):
            for file in files:
                file_path = str(os.path.join(root, file))

                if spec is not None and spec.match_file(file_path) is True:
                    continue

                file_size = os.path.getsize(file_path)

                if file_size > 10 * 1024 * 1024:  # check if file size is greater than 1MB
                    raise ValueError(f"File {file} is {file_size_to_mb(file_size, 2)} which is larger than 10MB")

                total_size += file_size
                if total_size > 64 * 1024 * 1024:  # check if total size is greater than 16MB
                    raise ValueError("The total size of the files is greater than 16MB")

                rel_path = os.path.relpath(str(os.path.join(root, file)), tar_dir)
                tar.add(str(os.path.join(root, file)), arcname=rel_path)

    base64_data = base64.b64encode(data.getvalue())

    base64_string = base64_data.decode('utf-8')

    return base64_string


def get_requirements_from_file(file_path: Path) -> List[str]:
    # TODO: refactor this duplicated code
    if file_path.exists() is False:
        return []

    print("Found requirements.txt using this")
    from pkg_resources import parse_requirements

    with open(file_path, 'r') as file:
        lines = file.readlines()
        # Ignore comments and empty lines
        lines = [line.strip() for line in lines if not line.startswith('#') and line.strip() != '']
        return [str(req) for req in parse_requirements(lines)]



import abc
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Sequence, Tuple, Literal, Dict, Union, List, Optional, Any, Type
from urllib.parse import urlencode

from databricks.sdk import WorkspaceClient
from databricks.sdk.service.serving import DataframeSplitInput

from mlrpc.proto import ResponseObject, RequestObject, HotReloadEvents, EncryptDecrypt, DataFileChunker

MethodType = Literal['GET', 'POST', 'PUT', 'DELETE', 'PATCH']
QueryParams = Dict[str, Union[str, List[str]]]
Headers = Sequence[Tuple[str, str]]
Body = Optional[str | Dict[str, Any] | List[Dict[str, Any]]]


@dataclass
class MLRPCResponse:
    request: RequestObject
    status_code: int
    headers: Optional[Sequence[Tuple[str, str]]] = None
    body: Optional[str | dict | list] = None

    @property
    def dict(self):
        if self.body is None:
            return {}
        if isinstance(self.body, dict):
            return self.body
        return json.loads(self.body)

    @property
    def json(self):
        return self.dict

    @property
    def text(self):
        if self.body is None:
            return ""
        if isinstance(self.body, str):
            return self.body
        return json.dumps(self.body)


def generate_query_string(params: QueryParams) -> str:
    return urlencode(params, doseq=True)


class DispatchHandler(abc.ABC):

    @abc.abstractmethod
    def dispatch(self, request: RequestObject) -> MLRPCResponse | List[MLRPCResponse]:
        pass


class MLRPCClient:
    def __init__(self, rpc_dispatch_handler: DispatchHandler):
        self._rpc_dispatch_handler: DispatchHandler = rpc_dispatch_handler

    def _dispatch(self,
                  *,
                  method: MethodType,
                  path: str,
                  query_params: Optional[QueryParams] = None,
                  data: Optional[Body] = None,
                  headers: Headers) -> MLRPCResponse:
        if query_params is not None:
            query_params = generate_query_string(query_params)
        if data is not None and isinstance(data, str) is False:
            data = json.dumps(data)
        request = RequestObject(
            method=method,
            path=path,
            query_params=query_params,
            headers=headers,
            content=data
        )
        resp = self._rpc_dispatch_handler.dispatch(request)
        if isinstance(resp, MLRPCResponse):
            return resp
        if isinstance(resp, list):
            return resp[0]

    def request(self,
                method: MethodType,
                path: str,
                query_params: Optional[QueryParams] = None,
                headers: Headers = None,
                data: Body = None,
                ) -> MLRPCResponse:
        return self._dispatch(method=method,
                              path=path,
                              data=data,
                              headers=headers,
                              query_params=query_params)

    def get(self,
            path: str,
            query_params: Optional[QueryParams] = None,
            headers: Headers = None,
            ) -> MLRPCResponse:
        return self._dispatch(method="GET",
                              path=path,
                              headers=headers,
                              query_params=query_params)

    def post(self,
             path: str,
             query_params: Optional[QueryParams] = None,
             data: Body = None,
             headers: Headers = None) -> MLRPCResponse:
        return self._dispatch(method="POST",
                              path=path,
                              headers=headers,
                              query_params=query_params,
                              data=data)

    def put(self, url,
            query_params: Optional[QueryParams] = None,
            data: Body = None,
            headers: Headers = None) -> MLRPCResponse:
        return self._dispatch(method="PUT",
                              path=url,
                              headers=headers,
                              query_params=query_params,
                              data=data)

    def patch(self, url,
              query_params: Optional[QueryParams] = None,
              data: Body = None,
              headers: Headers = None) -> MLRPCResponse:
        return self._dispatch(method="PATCH",
                              path=url,
                              headers=headers,
                              query_params=query_params,
                              data=data)

    def delete(self, url, headers: Headers = None) -> MLRPCResponse:
        return self._dispatch(method="DELETE",
                              path=url,
                              headers=headers)


class HotReloadMLRPCClient(MLRPCClient):

    def __init__(self, rpc_dispatch_handler: DispatchHandler):
        super().__init__(rpc_dispatch_handler)
        self._encrypt_decrypt = None

    def _setup_encryption(self):
        resp = self.get_public_key()[0]
        public_key = json.loads(resp.body)["public_key"]
        self._encrypt_decrypt = EncryptDecrypt(public_key=public_key)

    def hot_reload(self, directory_path) -> List[MLRPCResponse] | MLRPCResponse:
        if self._encrypt_decrypt is None:
            self._setup_encryption()
        from mlrpc.utils import dir_to_base64
        reload_dir = Path(directory_path)
        git_ignore = reload_dir / ".gitignore"
        content = dir_to_base64(reload_dir, git_ignore)
        return self._rpc_dispatch_handler.dispatch(HotReloadEvents.full_sync(content, self._encrypt_decrypt))

    def reinstall_requirements(self, requirements: List[str]) -> MLRPCResponse:
        return self._rpc_dispatch_handler.dispatch(HotReloadEvents.reinstall(requirements))

    def get_public_key(self) -> MLRPCResponse:
        return self._rpc_dispatch_handler.dispatch(HotReloadEvents.get_public_key())

    def upload_data_dir(self, root_dir: str, data_dir: str):
        for upload_file_req in DataFileChunker(root_dir=root_dir, data_dir=data_dir).iter_requests():
            self._rpc_dispatch_handler.dispatch(upload_file_req)


class LocalServingDispatchHandler(DispatchHandler):

    def __init__(self, host: str = "0.0.0.0", port: int = 5000, endpoint: str = "/invocations"):
        self._host = host
        self._port = port
        self._endpoint = endpoint

    def dispatch(self, request: RequestObject) -> MLRPCResponse | List[MLRPCResponse]:
        try:
            import httpx
        except ImportError:
            raise ImportError("httpx is required for LocalServingDispatchHandler; do pip install mlrpc[cli]")

        url = f"http://{self._host}:{self._port}{self._endpoint}"

        headers = request.headers or []
        headers.append(("Content-Type", "application/json"))
        # remove content-length header if present because the request is proxied and encoded in a custom way
        headers = [h for h in headers if h[0].lower() != "content-length"]
        content = request.encode().to_mlflow_df_split_dict()
        resp = httpx.request(method="POST", url=url, headers=headers,
                             json=content)
        predictions = resp.json()["predictions"]
        resp_objs = [ResponseObject.from_serving_resp(pred) for pred in predictions]
        return [MLRPCResponse(
            request=request,
            status_code=decoded_resp.status_code,
            headers=decoded_resp.headers,
            body=decoded_resp.content)
            for decoded_resp in resp_objs]


class ServingEndpointDispatchHandler(DispatchHandler):

    def __init__(self, endpoint_name: str, ws_client: WorkspaceClient = None):
        self._endpoint_name = endpoint_name
        self._ws = ws_client or WorkspaceClient()

    def dispatch(self, request: RequestObject) -> MLRPCResponse | List[MLRPCResponse]:
        serving_resp = self._ws.serving_endpoints.query(
            self._endpoint_name,
            dataframe_split=DataframeSplitInput(**request.encode().to_sdk_df_split())
        )
        decoded_resp = [ResponseObject.from_serving_resp(pred) for pred in serving_resp.predictions]
        return [MLRPCResponse(
            request=request,
            status_code=response.status_code,
            headers=response.headers,
            body=response.content
        )
            for response in decoded_resp]


class ServingRPCClient:

    def __init__(self, endpoint_name: str, ws_client: WorkspaceClient = None,
                 client_klass: Optional[Type[MLRPCClient] | Type[HotReloadMLRPCClient]] = None):
        client_klass = client_klass or MLRPCClient
        self._rpc_client = client_klass(ServingEndpointDispatchHandler(endpoint_name, ws_client))

    @property
    def rpc_client(self):
        return self._rpc_client


class LocalServingRPCClient:

    def __init__(self, host: str = "0.0.0.0", port: int = 6000, endpoint: str = "/invocations",
                 client_klass: Optional[Type[MLRPCClient] | Type[HotReloadMLRPCClient]] = None):
        client_klass = client_klass or MLRPCClient
        self._rpc_client = client_klass(LocalServingDispatchHandler(host, port, endpoint))

    @property
    def rpc_client(self):
        return self._rpc_client


class _Client:

    @staticmethod
    def local(host: str = "0.0.0.0", port: int = 6000, endpoint: str = "/invocations") -> MLRPCClient:
        return LocalServingRPCClient(host, port, endpoint).rpc_client

    @staticmethod
    def databricks(endpoint_name: str, ws_client: WorkspaceClient = None) -> MLRPCClient:
        return ServingRPCClient(endpoint_name, ws_client).rpc_client


class _HotReloadClient:

    @staticmethod
    def local(host: str = "0.0.0.0", port: int = 6000, endpoint: str = "/invocations") -> HotReloadMLRPCClient:
        return LocalServingRPCClient(host, port, endpoint, HotReloadMLRPCClient).rpc_client

    @staticmethod
    def databricks(endpoint_name: str, ws_client: WorkspaceClient = None) -> HotReloadMLRPCClient:
        return ServingRPCClient(endpoint_name, ws_client, HotReloadMLRPCClient).rpc_client


rpc = _Client()
hot_reload = _HotReloadClient()



import base64
import hashlib
import io
import json
import os
import shutil
import tarfile
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Sequence, Tuple, Dict, Literal, List, Callable

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa


def _b64encode(s: Optional[str]) -> Optional[str]:
    if s is None:
        return None
    return base64.b64encode(s.encode('utf-8')).decode('utf-8')


def _b64decode(s: Optional[str]) -> Optional[str]:
    if s is None:
        return None
    return base64.b64decode(s).decode('utf-8')


def _encode_anything(s: any) -> Optional[str]:
    if s is None:
        return None
    if isinstance(s, bytes):
        s = s.decode('utf-8')

    json_str = json.dumps(s)
    return _b64encode(json_str)


def _decode_anything(s: Optional[str]) -> any:
    if s is None:
        return None
    json_str = _b64decode(s)
    return json.loads(json_str)


def base64_to_dir(base64_string, target_dir,
                  skip_certain_directories: Optional[List[str]] = None,
                  debug_msg: Optional[Callable] = None):
    base64_data = base64.b64decode(base64_string)

    data = io.BytesIO(base64_data)
    tgt_dir_path = Path(target_dir)
    directory_names = [Path(skip_dir).name for skip_dir in skip_certain_directories or []]
    if debug_msg is not None:
        debug_msg(f"Skipping directories: {directory_names}", msg_type="HOT_RELOAD", level="INFO")
    for item in tgt_dir_path.iterdir():
        if item.is_dir() and item.name in (directory_names or []):
            continue  # Skip directories to ignore
        elif item.is_dir():
            shutil.rmtree(item)
        else:
            item.unlink()

    tgt_dir_path.mkdir(parents=True, exist_ok=True)

    with tarfile.open(fileobj=data, mode='r:gz') as tar:
        tar.extractall(path=target_dir)


@dataclass
class ResponseObjectEncoded:
    status_code: int
    headers: Optional[str] = None
    content: Optional[str] = None

    def dict(self):
        return {
            "status_code": self.status_code,
            "headers": self.headers,
            "content": self.content
        }

    def to_mlflow_df_split_dict(self):
        return {
            "dataframe_split": {
                "columns": ["status_code", "headers", "content"],
                "data": [[self.status_code, self.headers, self.content]],
            }
        }

    def to_mlflow_df_split_json(self):
        return json.dumps(self.to_mlflow_df_split_dict())


@dataclass
class ResponseObject:
    status_code: int
    headers: Optional[Sequence[Tuple[str, str]]] = None
    content: Optional[str] = None

    def encode(self):
        headers = [{header[0]: header[1]} for header in self.headers] if self.headers is not None else None
        return ResponseObjectEncoded(
            status_code=self.status_code,
            headers=_encode_anything(headers),
            content=_encode_anything(self.content)
        )

    @classmethod
    def from_resp_enc(cls, encoded: ResponseObjectEncoded):
        status_code = encoded.status_code
        headers = _decode_anything(encoded.headers) if encoded.headers is not None else []
        headers = [(k, v) for header in headers for k, v in header.items()]
        content = _decode_anything(encoded.content)
        return cls(
            status_code=status_code,
            headers=headers,
            content=content
        )

    @classmethod
    def from_serving_resp(cls, resp: Dict[str, str | int]):
        return cls(
            status_code=resp['status_code'],
            headers=_decode_anything(resp['headers']),
            content=_decode_anything(resp['content'])
        )

    @classmethod
    def from_httpx_resp(cls, resp: "httpx.Response"):
        # TODO: being lazy here, should handle more gracefully using mimetype
        try:
            return cls(
                status_code=resp.status_code,
                headers=list(resp.headers.items()),
                content=resp.json()
            )
        except Exception:
            return cls(
                status_code=resp.status_code,
                headers=list(resp.headers.items()),
                content=resp.text
            )


@dataclass
class RequestObjectEncoded:
    method: str
    path: str
    headers: Optional[str] = None
    query_params: Optional[str] = None
    content: Optional[str] = None
    timeout: Optional[float] = None

    def dict(self):
        return {
            "method": self.method,
            "path": self.path,
            "headers": self.headers,
            "query_params": self.query_params,
            "content": self.content,
            "timeout": self.timeout
        }

    def to_mlflow_df_split_dict(self):
        return {
            "dataframe_split": {
                "columns": ["method", "headers", "path", "query_params", "content", "timeout"],
                "data": [[self.method, self.headers, self.path, self.query_params, self.content, self.timeout]],
            }
        }

    def to_sdk_df_split(self):
        return {
            "columns": ["method", "headers", "path", "query_params", "content", "timeout"],
            "data": [[self.method, self.headers, self.path, self.query_params, self.content, self.timeout]],
        }

    def to_mlflow_df_split_json(self):
        return json.dumps(self.to_mlflow_df_split_dict())


@dataclass
class RequestObject:
    method: Literal['GET', 'POST', 'PUT', 'DELETE', 'PATCH']
    path: str
    headers: Optional[Sequence[Tuple[str, str]]] = None
    query_params: Optional[str] = None  # todo make this a bit more typed in future
    content: Optional[str] = None
    timeout: Optional[float] = None

    def encode(self):
        headers = [{header[0]: header[1]} for header in self.headers] if self.headers is not None else None
        return RequestObjectEncoded(
            method=self.method,
            headers=_encode_anything(headers),
            path=self.path,
            query_params=self.query_params,
            content=_encode_anything(self.content),
            timeout=self.timeout
        )

    @classmethod
    def from_request_enc(cls, encoded: RequestObjectEncoded):
        method: Literal['GET', 'POST', 'PUT', 'DELETE', 'PATCH'] = encoded.method
        headers = _decode_anything(encoded.headers) if encoded.headers is not None else []
        headers = [(k, v) for header in headers for k, v in header.items()]
        path = encoded.path
        query_params = encoded.query_params
        content = _decode_anything(encoded.content)
        timeout = encoded.timeout
        return cls(
            method=method,
            headers=headers,
            path=path,
            query_params=query_params,
            content=content,
            timeout=timeout
        )


# HOT RELOAD EVENTS ARE ALWAYS POSTS
class HotReloadEvents:

    @staticmethod
    def full_sync(content: str, encryption_key: "EncryptDecrypt") -> RequestObject:
        enc_content = encryption_key.encrypt(content)
        return RequestObject(
            method="POST",
            path="/__INTERNAL__/FULL_SYNC",
            content=json.dumps({
                "content": enc_content,
                "checksum": hashlib.md5(content.encode('utf-8')).hexdigest()
            })
        )

    @staticmethod
    def upload_large_file(c: "Chunk"):
        return RequestObject(
            method="POST",
            path="/__INTERNAL__/UPLOAD_LARGE_FILE",
            content=json.dumps({
                "chunk": c.dict()
            })
        )

    @staticmethod
    def reload() -> RequestObject:
        return RequestObject(
            method="POST",
            path="/__INTERNAL__/RELOAD"
        )

    @staticmethod
    def get_public_key() -> RequestObject:
        return RequestObject(
            method="POST",
            path="/__INTERNAL__/GET_PUBLIC_KEY"
        )

    @staticmethod
    def reinstall(requirements: List[str]) -> RequestObject:
        return RequestObject(
            method="POST",
            path="/__INTERNAL__/REINSTALL",
            content=json.dumps({
                "requirements": requirements
            })
        )


class KeyGenerator:
    def __init__(self, directory, debug_msg: Optional[Callable] = None):
        self.directory = directory
        self.private_key = None
        self.public_key = None
        self._debug_msg = debug_msg or print
        self._public_key_name = 'public_key.pem'
        self._private_key_name = 'private_key.pem'

    def generate(self):
        private_key_path = os.path.join(self.directory, self._private_key_name)
        public_key_path = os.path.join(self.directory, self._public_key_name)

        if os.path.exists(private_key_path) and os.path.exists(public_key_path):
            self._debug_msg(f"Keys already exist in {self.directory}")
            return

        self.private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )
        self.public_key = self.private_key.public_key()

        private_key_pem = self.private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        )
        public_key_pem = self.public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        with open(private_key_path, 'wb') as f:
            f.write(private_key_pem)

        with open(public_key_path, 'wb') as f:
            f.write(public_key_pem)

        self._debug_msg(f"Keys generated and saved to {self.directory}")

    def _get_or_regenerate(self, key_name):
        if os.path.exists(os.path.join(self.directory, key_name)) is False:
            self._debug_msg(f"{key_name} does not exist in {self.directory} so regenerating...")
            self.generate()
        with open(os.path.join(self.directory, key_name), 'rb') as f:
            return f.read().decode()

    def get_public_key(self):
        return self._get_or_regenerate(self._public_key_name)

    def get_private_key(self):
        return self._get_or_regenerate(self._private_key_name)


class EncryptDecrypt:
    def __init__(self, *,
                 private_key=None,
                 public_key=None,
                 key_generator: Optional[KeyGenerator] = None):
        self.private_key = serialization.load_pem_private_key(
            private_key.encode(),
            password=None
        ) if private_key else None
        self.public_key = serialization.load_pem_public_key(
            public_key.encode()
        ) if public_key else None
        self.key_generator = key_generator
        self.chunk_delimiter = '  '

    def encrypt(self, message: str) -> str:
        if not self.public_key and self.key_generator is not None:
            public_key = self.key_generator.get_public_key()
            self.public_key = serialization.load_pem_public_key(
                public_key.encode()
            )
        if not self.public_key:
            raise ValueError("Public key is not set")

        # magic number for 2048 bit RSA key
        chunk_size = 150

        # encrypt chunks
        encrypted_chunks = []
        for i in range(0, len(message), chunk_size):
            chunk = message[i:i + chunk_size]
            encrypted_chunk = self.public_key.encrypt(
                chunk.encode(),
                padding.OAEP(
                    mgf=padding.MGF1(algorithm=hashes.SHA256()),
                    algorithm=hashes.SHA256(),
                    label=None
                )
            )
            encrypted_chunks.append(base64.b64encode(encrypted_chunk).decode())

        encrypted_content = self.chunk_delimiter.join(encrypted_chunks)
        return encrypted_content

    def decrypt(self, b64_ciphertext: str):
        if not self.private_key and self.key_generator is not None:
            private_key = self.key_generator.get_private_key()
            self.private_key = serialization.load_pem_private_key(
                private_key.encode(),
                password=None
            )

        if not self.private_key:
            raise ValueError("Private key is not set")

        encrypted_chunks = b64_ciphertext.split(self.chunk_delimiter)
        decrypted_chunks = []

        for chunk in encrypted_chunks:
            ciphertext = base64.b64decode(chunk)
            plaintext = self.private_key.decrypt(
                ciphertext,
                padding.OAEP(
                    mgf=padding.MGF1(algorithm=hashes.SHA256()),
                    algorithm=hashes.SHA256(),
                    label=None
                )
            )
            decrypted_chunks.append(plaintext.decode())

        return ''.join(decrypted_chunks)


@dataclass
class Chunk:
    content_b64: str
    chunk_number: int
    next_offset: int
    encrypted: bool = False
    eof: bool = False
    start_of_file: bool = False
    permissions: Optional[str] = None
    relative_file_path: Optional[str] = None

    # TODO: include checksums

    def dict(self):
        return {
            "content_b64": self.content_b64,
            "chunk_number": self.chunk_number,
            "next_offset": self.next_offset,
            "encrypted": self.encrypted,
            "eof": self.eof,
            "start_of_file": self.start_of_file,
            "permissions": self.permissions,
            "relative_file_path": self.relative_file_path
        }

    def __str__(self):
        return (f"Chunk {self.chunk_number} - {self.next_offset} - {self.encrypted} - "
                f"{self.eof} - {self.start_of_file} - {self.permissions} - {self.relative_file_path}")

    def __repr__(self):
        return self.__str__()


class DataFileChunker:

    def __init__(self,
                 root_dir: str,
                 data_dir: str,
                 chunk_size_bytes: int = 10000000,
                 encrypt_decrypt: Optional[EncryptDecrypt] = None):
        self.data_dir = data_dir
        self.root_dir = root_dir
        self.chunk_size_bytes = chunk_size_bytes
        self.encrypt_decrypt = encrypt_decrypt

    def _get_offset(self, last_chunk: Optional[Chunk] = None) -> int:
        if last_chunk is None:
            return 0

        return last_chunk.next_offset or 0

    def _is_eof(self, last_chunk: Optional[Chunk] = None) -> bool:
        if last_chunk is None:
            return False
        return last_chunk.eof

    def make_chunk(self, file_path: str, last_chunk: Optional[Chunk] = None) -> Chunk:
        with open(file_path, 'rb') as f:
            offset = self._get_offset(last_chunk)
            f.seek(offset)
            content = f.read(self.chunk_size_bytes)
            content_b64 = base64.b64encode(content).decode()
            if self.encrypt_decrypt:
                content_b64 = self.encrypt_decrypt.encrypt(content_b64)

            return Chunk(
                relative_file_path=Path(file_path).relative_to(self.root_dir).as_posix(),
                start_of_file=offset == 0,
                permissions=oct(os.stat(file_path).st_mode & 0o777),
                content_b64=content_b64,
                chunk_number=last_chunk.chunk_number + 1 if last_chunk else 0,
                next_offset=f.tell(),
                encrypted=bool(self.encrypt_decrypt),
                eof=f.tell() == os.fstat(f.fileno()).st_size
            )

    def make_n_chunks(self, file_path: str, n: int, last_chunk: Optional[Chunk] = None) -> List[Chunk]:
        chunks = []
        for _ in range(n):
            if self._is_eof(last_chunk):
                break
            chunk = self.make_chunk(file_path, last_chunk=last_chunk)
            chunks.append(chunk)
            last_chunk = chunk
        return chunks

    def iter_chunks(self, file_path: str, last_chunk: Optional[Chunk] = None):
        while True:
            if self._is_eof(last_chunk):
                break
            chunk = self.make_chunk(file_path, last_chunk=last_chunk)
            yield chunk
            last_chunk = chunk

    def iter_files_chunks(self):
        root_path = Path(self.root_dir)
        data_path = root_path / self.data_dir
        for file in data_path.rglob("*"):
            for chunk in self.iter_chunks(str(file)):
                yield chunk

    def iter_requests(self):
        for chunk in self.iter_files_chunks():
            yield HotReloadEvents.upload_large_file(c=chunk)


class DataFileChunkWriter:

    def __init__(self,
                 root_dir: str,
                 encrypt_decrypt: Optional[EncryptDecrypt] = None):
        self.root_dir = root_dir
        self.encrypt_decrypt = encrypt_decrypt

    def write_chunk(self, chunk: Chunk):
        target_file = Path(self.root_dir) / chunk.relative_file_path
        target_file.parent.mkdir(parents=True, exist_ok=True)
        write_mode = 'wb' if chunk.start_of_file else 'ab'
        with target_file.open(write_mode) as f:
            content = base64.b64decode(chunk.content_b64)
            if chunk.encrypted:
                decrypted = self.encrypt_decrypt.decrypt(chunk.content_b64)
                content: bytes = base64.b64decode(decrypted)
            f.write(content)
            f.flush()
            os.fsync(f.fileno())
        if chunk.permissions is not None:
            os.chmod(target_file, int(chunk.permissions, 8))
        return True

    def write_chunks(self, chunks: List[Chunk]):
        for chunk in chunks:
            self.write_chunk(chunk)
        return True



import os
import sys
import tempfile
import threading
import time
import webbrowser
from pathlib import Path
from subprocess import CalledProcessError
from typing import Dict, Optional, Callable
from urllib.parse import urlparse

import click
import uvicorn
from databricks.sdk import WorkspaceClient

from mlrpc.cfg import ConfigFileProcessor, INIT_CONFIG
from mlrpc.client import hot_reload
from mlrpc.deployment import get_or_create_mlflow_experiment, save_model, keep_only_last_n_versions, \
    deploy_secret_env_file, deploy_serving_endpoint, default_mlrpc_libs, stage_files_for_deployment, copy_files
from mlrpc.detect import scan_in_directory
from mlrpc.flavor import FastAPIFlavor, pack_env_file_into_preload
from mlrpc.hotreload import hot_reload_on_change, make_log_monitor_thread
from mlrpc.proxy import make_swagger_proxy
from mlrpc.utils import execute, find_next_open_port, get_profile_contents, DatabricksProfile, get_version


@click.group()
def cli():
    """
    A CLI for deploying services to databricks model serving
    """
    pass


def configure_mlflow_to_databricks(env_dict: Dict[str, str], profile: Optional[DatabricksProfile] = None):
    env_dict["MLFLOW_TRACKING_URI"] = "databricks"
    env_dict["MLFLOW_REGISTRY_URI"] = "databricks-uc"
    if profile is not None:
        env_dict["DATABRICKS_HOST"] = profile.host
        env_dict["DATABRICKS_TOKEN"] = profile.token
    return env_dict


def serve_mlflow_model_cmd(model_uri: str, port: int):
    return [
        "mlflow",
        "models",
        "serve",
        "-h",
        "0.0.0.0",
        "-m",
        model_uri,
        "-w",  # mimic production
        "4",
        "-p",
        str(port),
        # "--env-manager",  # for local testing
        # "local",  # for local testing
    ]


def get_only_host(url: str):
    return urlparse(url).hostname


def get_catalog_url(host: str, uc_name: str, version: str):
    catalog, schema, model = uc_name.split(".")
    return f"https://{host}/explore/data/models/{catalog}/{schema}/{model}/version/{version}"


def get_experiment_url(host: str, experiment_id: str):
    return f"https://{host}/ml/experiments/{experiment_id}"


def get_endpoint_events_url(host: str, endpoint_name: str):
    return f"https://{host}/ml/endpoints/{endpoint_name}/events"


CONTEXT_SETTINGS = dict(default_map=ConfigFileProcessor.read_config())


def ensure_run_or_uc(run_name: str, catalog, schema, name, latest_alias_name):
    if run_name is None and (catalog is None or schema is None or name is None or latest_alias_name is None):
        raise click.ClickException("Either provide a run name or a catalog, schema and model name")
    if run_name is not None and (
            catalog is not None or schema is not None or name is not None or latest_alias_name is not None):
        raise click.ClickException("Provide either a run name or a catalog, schema and model name")


def ensure_databrickscfg_exists():
    if not Path("~/.databrickscfg").expanduser().exists():
        raise click.ClickException("No databrickscfg file found in ~/.databrickscfg. Please make a profile")


def make_full_uc_path(catalog: str, schema: str, name: str):
    return f"{catalog}.{schema}.{name}"


@cli.command("version")
def _version():
    """
    Print the version of the package
    """
    version = get_version()
    click.echo(f"Version: {version}\n")

    click.echo("Listing mlrpc default versions:")
    for idx, lib in enumerate(default_mlrpc_libs()):
        click.echo(f"  {idx + 1}. {lib}")


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.option("-c", "--catalog", "uc_catalog", type=str, required=True,
              help="The unity catalog name of the model")
@click.option("-s", "--schema", "uc_schema", type=str, required=True,
              help="The unity schema name of the model")
@click.option("-n", "--name", "name", type=str, required=True,
              help="The name of the app you want to deploy")
@click.option("-a", "--alias", "latest_alias_name", type=str, required=True,
              help="The alias name of the model that will be deployed")
@click.option("-r", "--run-name", "run_name", type=str,
              help="The name of the run to deploy")
@click.option("-p", "--profile", "databricks_profile", type=str, default=None,
              help="The databricks profile to use. This is the section name in ~/.databrickscfg file.")
@click.option("-e", "--env-file", "envfile", type=click.Path(exists=True, resolve_path=True, file_okay=True),
              default=None, help="The location of the env file to deploy")
@click.option("-h", "--headless", "headless", is_flag=True, default=False,
              help="Run local swagger server without opening browser")
@click.option("--no-swagger", "no_swagger", is_flag=True, default=False,
              help="Do not run swagger")
@click.pass_context
def local(
        ctx,
        uc_catalog: str,
        uc_schema: str,
        name: str,
        run_name: str,
        latest_alias_name: str,
        databricks_profile: str,
        envfile: str,
        headless: bool,
        no_swagger: bool,
):
    """
    Serve a model locally from remote databricks model registry
    """
    ensure_run_or_uc(run_name, uc_catalog, uc_schema, name, latest_alias_name)
    ensure_databrickscfg_exists()
    profile = get_profile_contents(databricks_profile)
    env_copy = os.environ.copy()
    env_copy = configure_mlflow_to_databricks(env_copy, profile)
    proxy_server_port = find_next_open_port(6500, 7000)
    mlflow_server_port = find_next_open_port(8000, 9000)
    ws = WorkspaceClient(profile=databricks_profile)
    # TODO: this is a hack with some weird boto3 bug if you try to directly access the model version
    if envfile is not None:
        pack_env_file_into_preload(Path(envfile), env_copy)
    uc_name = make_full_uc_path(uc_catalog, uc_schema, name)
    alias = latest_alias_name + "-devel"
    v = ws.model_versions.get_by_alias(uc_name, alias)
    host = get_only_host(profile.host)
    click.echo(click.style(f"Model URL: {get_catalog_url(host, uc_name, str(v.version))}", fg="green"))
    click.echo("\n")

    def start_swagger():
        app = make_swagger_proxy("",
                                 profile=databricks_profile,
                                 port=proxy_server_port,
                                 databricks_mode=False,
                                 local_server_port=mlflow_server_port)
        thread, cb = swagger_in_thread(app, proxy_server_port, headless=headless)
        click.echo(click.style(f"Swagger UI available at: http://0.0.0.0:{proxy_server_port}/docs", fg="green"))
        return thread, cb

    swagger_thread = None
    count = 0
    try:
        for log in execute(
                cmd=serve_mlflow_model_cmd(
                    f"runs:/{v.run_id}/model"
                    , mlflow_server_port),
                env=env_copy,
        ):
            if "[INFO] Listening at:" in log and no_swagger is False and count == 0:
                click.clear()
                click.echo(log)
                swagger_thread, open_browser_cb = start_swagger()
                swagger_thread.start()
                open_browser_cb()
                hot_reload_on_change(Path.cwd(), hot_reload.local(port=mlflow_server_port),
                                     logging_function=lambda x: click.echo(click.style(x, fg="yellow")),
                                     error_logging_function=lambda x: click.echo(click.style(x, fg="red", bold=True)),
                                     success_logging_function=lambda x: click.echo(
                                         click.style(x, fg="green", bold=True)))
                count += 1
            else:
                click.echo(log)
    except CalledProcessError as e:
        raise click.ClickException("Error serving model")

    if swagger_thread is not None:
        swagger_thread.join()


def success_scanning_for_issues(directory: str) -> bool:
    click.echo(click.style("Scanning for secrets in staged files", fg="green"))
    issues = list(scan_in_directory(str(directory)))
    # first check if issues have something
    if issues is not None and len(issues) > 0:
        # then filter out the None issues
        issues_filtered = [i for i in issues if i is not None]
        # then check if the filtered issues have something
        if issues_filtered is not None and len(issues_filtered) > 0:
            click.echo(click.style("Issues found in staged files", fg="yellow"))
            for issue in issues_filtered:
                click.echo(click.style(issue, fg="red", bold=True))
            if sys.stdin.isatty() is False or sys.stdout.isatty() is False or sys.stderr.isatty() is False:
                raise click.ClickException("Issues found in staged files and stdout is not a tty. Aborting deployment.")
            if click.confirm(click.style(
                    "We found potential secrets in the previous statement, would you like to continue?",
                    fg="red",
                    bold=True
            )):
                click.echo(click.style("Continuing with deployment", fg="green"))
                return True
            else:
                click.echo(click.style("Aborting deployment", fg="red"))
                return False
    click.echo(click.style("No issues found in staged files", fg="green"))
    return True


def apply_build_options(f: Callable):
    options = [
        click.option("-c", "--catalog", "uc_catalog", type=str, required=True,
                     help="The unity catalog name of the model"),
        click.option("-s", "--schema", "uc_schema", type=str, required=True,
                     help="The unity schema name of the model"),
        click.option("--app-root-dir", "app_root_dir", required=True,
                     type=click.Path(exists=True, file_okay=False, dir_okay=True),
                     help="The root directory of the app"),
        click.option("--app-path-in-root-dir", "app_path_in_root", type=str, required=True,
                     help="The path to the app in the root directory"),
        click.option("--app-object", "app_obj", type=str, required=True,
                     help="The name of the app object in the app file"),
        click.option("--data-dir", "data_dir", default=None,
                     type=click.Path(exists=True, resolve_path=True, file_okay=False, dir_okay=True)),
        click.option("-n", "--name", "name", type=str, required=True,
                     help="The name of the app you want to deploy"),
        click.option("-a", "--alias", "latest_alias_name", type=str, required=True,
                     help="The alias name of the model that will be deployed"),
        click.option("--make-experiment", "make_experiment", type=bool, default=True,
                     help="Whether to create a new experiment"),
        click.option("--experiment-name", "experiment_name", type=str, default=None,
                     help="The name of the experiment to create"),
        click.option("-r", "--register-model", "register_model", type=bool, default=True,
                     help="Whether to register the model"),
        click.option("-p", "--databricks-profile", "databricks_profile", type=str, default=None,
                     help="The databricks profile to use. This is the section name in ~/.databrickscfg file."),
        click.option("-e", "--env", "env", type=str, default=None,
                     help="The environment to deploy the api to"),
        click.option("--only-last-n-versions", "only_last_n_versions", type=int, default=None,
                     help="The number of versions to keep"),
        click.option("--bootstrap-python-script", "bootstrap_python_script",
                     type=click.Path(exists=True, file_okay=True), default=None,
                     help="The bootstrap script to run on each container"),
    ]
    for option in options:
        f = option(f)
    return f


@cli.command(context_settings=CONTEXT_SETTINGS)
@apply_build_options
@click.pass_context
def build(ctx, *,
          uc_catalog: str,
          uc_schema: str,
          app_root_dir: str,
          app_path_in_root: str,
          app_obj: str,
          data_dir: str,
          name: str,
          latest_alias_name: str,
          make_experiment: bool,
          experiment_name: str,
          register_model: bool,
          databricks_profile: str,
          only_last_n_versions: int,
          bootstrap_python_script: str,
          env: str,
          ):
    """
    Build a model and save to databricks model registry
    """
    ensure_databrickscfg_exists()
    if databricks_profile is not None:
        profile = get_profile_contents(databricks_profile)
        configure_mlflow_to_databricks(os.environ, profile)
        host = get_only_host(profile.host)
    else:
        host = get_only_host(os.environ["DATABRICKS_HOST"])

    ws = WorkspaceClient(profile=databricks_profile)
    uc_name = make_full_uc_path(uc_catalog, uc_schema, name)
    generated_experiment_name = f"{uc_catalog}_{uc_schema}_{name}"
    prod_model = FastAPIFlavor(local_app_dir_abs=str(app_root_dir),
                               local_app_path_in_dir=app_path_in_root,
                               app_obj=app_obj, reloadable=False, data_dir=data_dir,
                               bootstrap_script=bootstrap_python_script)
    devel_model = FastAPIFlavor(local_app_dir_abs=str(app_root_dir),
                                local_app_path_in_dir=app_path_in_root,
                                app_obj=app_obj, reloadable=True, data_dir=data_dir,
                                bootstrap_script=bootstrap_python_script)
    if make_experiment is True:
        created_experiment_name = generated_experiment_name if experiment_name is None else experiment_name
        exp = get_or_create_mlflow_experiment(ws, created_experiment_name)
        click.echo(click.style(f"Created experiment {created_experiment_name}", fg="green"))
        click.echo(click.style(f"Experiment URL: {get_experiment_url(host, exp.experiment_id)}", fg="green"))
    else:
        created_experiment_name = generated_experiment_name if experiment_name is None else experiment_name
        exp = ws.experiments.get_by_name(created_experiment_name)
        click.echo(click.style(f"Using experiment {uc_name}", fg="green"))
        click.echo(click.style(f"Experiment URL: {get_experiment_url(host, exp.experiment.experiment_id)}", fg="green"))

    if register_model is True:
        click.echo(click.style("Registering Application", fg="green"))
        with tempfile.TemporaryDirectory() as temp_dir:
            click.echo(click.style(f"Staging files for deployment in: {temp_dir}", fg="green"))
            stage_files_for_deployment(prod_model, str(temp_dir))
            if data_dir is not None:
                relative_path = Path(data_dir).relative_to(app_root_dir)
                target_dir = Path(temp_dir) / relative_path
                click.echo(click.style(f"Copying data directory: {data_dir} to: {target_dir}", fg="green"))
                copy_files(data_dir, target_dir, None)
            click.echo(click.style(f"Staged files for deployment in: {temp_dir}", fg="green"))
            status = success_scanning_for_issues(temp_dir)
            if status is False:
                return
            click.echo(click.style("Registering Development app", fg="green"))
            model_version = save_model(ws_client=ws,
                                       experiment=exp,
                                       app=devel_model,
                                       uc_model_path=uc_name,
                                       code_path=str(temp_dir),
                                       latest_alias_name=latest_alias_name + "-devel",
                                       reload=True)
            click.echo(
                click.style(
                    f"Development App Artifacts URL: {get_catalog_url(host, uc_name, str(model_version.version))}",
                    fg="green"))
            click.echo(click.style("Registering Production app", fg="green"))
            model_version = save_model(ws_client=ws,
                                       experiment=exp,
                                       app=prod_model,
                                       uc_model_path=uc_name,
                                       code_path=str(temp_dir),
                                       reload=False)
            click.echo(
                click.style(
                    f"Production App Artifacts URL: {get_catalog_url(host, uc_name, str(model_version.version))}",
                    fg="green"))
            if only_last_n_versions is not None and only_last_n_versions > 1:
                click.echo(click.style(f"Only last {only_last_n_versions} versions will be kept", fg="green"))
                keep_only_last_n_versions(ws, uc_name, only_last_n_versions)
    else:
        click.echo(click.style("Skipping model registration", fg="yellow"))


@cli.command("deploy", context_settings=CONTEXT_SETTINGS, deprecated=True, hidden=True)
@apply_build_options
@click.pass_context
def deploy_deprecated(ctx, *args, **kwargs):
    """Use `build` instead. This command will be removed in future releases!"""
    click.echo('WARNING: `old_command` is deprecated, use `build` instead.')
    build(ctx, *args, **kwargs)


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.option("-c", "--catalog", "uc_catalog", type=str, required=True,
              help="The unity catalog name of the model")
@click.option("-s", "--schema", "uc_schema", type=str, required=True,
              help="The unity schema name of the model")
@click.option("-n", "--name", "name", type=str, required=True,
              help="The name of the app you want to deploy")
@click.option("-a", "--alias", "latest_alias_name", type=str, required=True,
              help="The alias name of the model that will be deployed")
@click.option("-e", "--env", "env", type=str, default=None)
@click.option("--endpoint-name", "endpoint_name", type=str, required=True,
              help="The name of the endpoint to deploy")
@click.option("--secret-scope", "secret_scope", type=str, default=None,
              help="The secret scope to deploy the env file to")
@click.option("--secret-key", "secret_key", type=str, default=None,
              help="The secret key to deploy the env file to")
@click.option("--env-file", "env_file", type=click.Path(exists=True, resolve_path=True, file_okay=True),
              default=None, help="The location of the env file to deploy")
@click.option("-p", "--databricks-profile", "databricks_profile", type=str, default=None,
              help="The databricks profile to use. This is the section name in ~/.databrickscfg file.")
@click.option("--size", "size", type=str, default="Small",
              help="The size of the instance to deploy the endpoint to")
@click.option("--type", "type", type=str, default="CPU",
              help="The type of workload (CPU, or various GPU Sizes)")
@click.option("--scale-to-zero-enabled", "scale_to_zero_enabled", type=bool, default=True,
              help="Whether to enable scale to zero for the endpoint")
@click.option("--prod", "prod", is_flag=True, default=False,
              help="Enable production app deployment, disables reload server")
@click.pass_context
def serve(ctx, *,
          uc_catalog: str,
          uc_schema: str,
          name: str,
          latest_alias_name: str,
          env: str,
          endpoint_name: str,
          databricks_profile: str,
          secret_scope: str,
          secret_key: str,
          env_file: str,
          size: str,
          scale_to_zero_enabled: bool,
          prod: bool,
          type: str,  # noqa
          ):
    """
    Deploy a serving endpoint to databricks model serving
    """
    ws = WorkspaceClient(profile=databricks_profile)
    uc_name = make_full_uc_path(uc_catalog, uc_schema, name)
    if databricks_profile is not None:
        profile = get_profile_contents(databricks_profile)
        configure_mlflow_to_databricks(os.environ, profile)
        host = get_only_host(profile.host)
    else:
        host = get_only_host(os.environ["DATABRICKS_HOST"])

    click.echo(click.style(
        f"Deploying serving endpoint: {endpoint_name} for model: {uc_name} with version: {latest_alias_name}",
        fg="green"))
    if env_file is not None:
        click.echo(
            click.style(f"Deploying secret env file to secret scope: {secret_scope} and key: {secret_key}", fg="green"))
        if secret_scope is None or secret_key is None:
            raise click.ClickException("Both secret scope and key must be provided to deploy env file")
        deploy_secret_env_file(ws_client=ws,
                               secret_scope=secret_scope,
                               secret_key=secret_key,
                               env_file=Path(env_file))

    version = ws.model_versions.get_by_alias(uc_name, latest_alias_name)
    reload_version = ws.model_versions.get_by_alias(uc_name, latest_alias_name + "-devel") if prod is False else None

    if endpoint_name is None:
        raise click.ClickException("Endpoint name must be provided")

    events_url = get_endpoint_events_url(host, endpoint_name)
    click.echo(click.style(f"Follow Endpoint Events at URL: {events_url}", fg="green"))

    deploy_serving_endpoint(ws_client=ws,
                            endpoint_name=endpoint_name,
                            uc_model_path=uc_name,
                            model_version=version,
                            reload_version=reload_version,
                            workload_type=type,
                            scale_to_zero_enabled=scale_to_zero_enabled,
                            secret_key=secret_key,
                            secret_scope=secret_scope,
                            size=size)


def make_hotreload_threads(ws_client: WorkspaceClient,
                           endpoint_name: str,
                           databricks_profile: str,
                           debug: bool,
                           headless: bool,
                           hotreload_port: int):
    endpoint_suffix = "/served-models/hotreload"
    print("Starting hot reload server")
    hotreload_app = make_swagger_proxy(endpoint_name + endpoint_suffix, profile=databricks_profile,
                                       port=hotreload_port,
                                       debug=debug, hotreload_enabled=True)
    hotreload_swagger_thread, hotreload_browser_cb = swagger_in_thread(
        hotreload_app, hotreload_port, headless=headless, proxy_name="HOTRELOAD-PROXY")
    hotreload_swagger_thread.start()
    hotreload_browser_cb()
    rpc_client = hot_reload.databricks(endpoint_name=endpoint_name + endpoint_suffix,
                                       ws_client=ws_client)
    pwd = Path.cwd()
    prefix = click.style("[HOT-RELOADER]", fg="blue", bold=True) + ' '
    click.echo(prefix + click.style(f"Doing full sync of: {str(pwd)}", fg="green"))
    rpc_client.hot_reload(str(pwd))
    reload_threads = hot_reload_on_change(pwd,
                                          rpc_client=rpc_client,
                                          logging_function=lambda x: click.echo(prefix + x),
                                          error_logging_function=lambda x: click.echo(
                                              prefix + click.style(x, fg="red", bold=True)),
                                          success_logging_function=lambda x: click.echo(
                                              prefix + click.style(x, fg="green", bold=True)))
    reload_threads.append(hotreload_swagger_thread)
    return reload_threads


def apply_connect_options(f: Callable):
    options = [
        click.option("--endpoint-name", "endpoint_name",
                     required=True,
                     help="The name of the databricks endpoint to explore",
                     type=str),
        click.option("-p", "--databricks-profile", "databricks_profile",
                     help="The databricks profile to use. This is the section name in ~/.databrickscfg file.",
                     type=str,
                     default=None),
        click.option("-d", "--debug", "debug", is_flag=True, default=False),
        click.option("-h", "--headless", "headless", is_flag=True, default=False,
                     help="Run local swagger server without opening browser"),
        click.option("-r", "--reload", "reload", is_flag=True, default=False,
                     help="Reload remotely deployed endpoint")
    ]
    for option in options:
        f = option(f)
    return f


@cli.command(context_settings=CONTEXT_SETTINGS)
@apply_connect_options
@click.pass_context
def connect(
        ctx,
        endpoint_name: str,
        databricks_profile: str,
        debug: bool,
        headless: bool,
        reload: bool
):
    """
    Explore a databricks endpoint using a swagger UI
    """

    # run the main app
    main_port = find_next_open_port(8000, 8499)
    hotreload_port = find_next_open_port(8500, 9000)
    app = make_swagger_proxy(endpoint_name, profile=databricks_profile, port=main_port, debug=debug)
    click.clear()
    click.echo(click.style(f"Swagger UI available at: http://0.0.0.0:{main_port}/docs", fg="green"))
    if reload is True:
        click.echo(click.style(f"Hot Reload Swagger UI available at: http://0.0.0.0:{hotreload_port}/docs",
                               fg="green", bold=True))
    thread, browser_cb = swagger_in_thread(app, main_port, headless=headless)
    click.echo("\n\n")
    thread.start()
    browser_cb()
    ws_client = WorkspaceClient(profile=databricks_profile)
    logging_thread = make_log_monitor_thread(
        ws_client,
        endpoint_name,
        logging_function=lambda x: click.echo(click.style("[MODEL-SERVING] ", fg="cyan", bold=True) + f"{x}"),
    )
    logging_thread.start()
    reload_threads = None
    if reload is True:
        # run the reload threads
        reload_threads = make_hotreload_threads(ws_client,
                                                endpoint_name,
                                                databricks_profile,
                                                debug,
                                                headless,
                                                hotreload_port)

    thread.join()
    if reload_threads is not None:
        for t in reload_threads:
            t.join()

    logging_thread.join()


@cli.command("swagger", context_settings=CONTEXT_SETTINGS, deprecated=True, hidden=True)
@apply_connect_options
@click.pass_context
def swagger_deprecated(ctx, *args, **kwargs):
    """Use `connect` instead. This command will be removed in future releases!"""
    connect(ctx, *args, **kwargs)


@cli.command()
def init():
    """
    Create a default config file
    """
    valid_cfgs = ConfigFileProcessor.config_files
    paths = [Path(cfg) for cfg in valid_cfgs]
    if any([p.exists() for p in paths]):
        raise click.ClickException(f"Config file already exists. {paths}")
    click.echo("Initializing mlrpc")
    Path("mlrpc.cfg").write_text(INIT_CONFIG)
    click.echo("Config file created at mlrpc.cfg")


def swagger_in_thread(app, port: int, host: str = "0.0.0.0", headless: bool = False, proxy_name: str = "MLRPC-PROXY"):
    def run_server():
        log_config = uvicorn.config.LOGGING_CONFIG
        prefix = click.style(f"[{proxy_name}]", fg="magenta", bold=True)
        log_formatter = f"{prefix} [%(asctime)s] - [%(levelname)s] - %(message)s"
        log_config["formatters"]["access"]["fmt"] = log_formatter
        log_config["formatters"]["default"]["fmt"] = log_formatter
        uvicorn.run(app, host=host, port=port, log_config=log_config)

    server_thread = threading.Thread(target=run_server, daemon=True)

    def browser_callback():
        if headless is False:
            time.sleep(0.5)
            webbrowser.open(f"http://{host}:{port}/docs")

    return server_thread, browser_callback



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="mlrpc",
    author="Sri Tikkireddy",
    author_email="sri.tikkireddy@databricks.com",
    description="Deploy FastAPI applications on MLFlow",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_data={
        'mlrpc': ['**/*.html'],
    },
    url="https://github.com/stikkireddy/mlflow-rpc",
    packages=find_packages(),
    install_requires=[],
    setup_requires=["setuptools_scm"],
    extras_require={
        "cli": [
            "pathspec",
            "mlflow-skinny[databricks]",
            "scipy", # only required to deploy model due to infer schema for inferring matrices
            "fastapi",
            "pandas",
            "httpx",
            "databricks-sdk",
            "click",
            "click-configfile",
            "python-dotenv",
            "uvicorn",
            "virtualenv",
            "flask", # only required for testing mlflow locally
            "watchdog",
            "detect-secrets",
        ],
        "client": [
            "databricks-sdk",
        ]
    },
    use_scm_version=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'mlrpc = mlrpc.cli:cli',
        ],
    },
    python_requires=">=3.10",
)



import queue
import re
import threading
import time
from datetime import datetime
from pathlib import Path
from typing import Callable, Literal, List
from typing import Optional

from databricks.sdk import WorkspaceClient
from dateutil.parser import parse
from pathspec import PathSpec
from pathspec.patterns import GitWildMatchPattern
from pytz import timezone
from watchdog.events import FileSystemEventHandler, FileSystemEvent
from watchdog.observers import Observer

from mlrpc.client import MLRPCResponse, HotReloadMLRPCClient
from mlrpc.utils import get_requirements_from_file


def get_ignore_specs_for_hotreload(dir_to_watch) -> Optional[PathSpec]:
    ignore_file = Path(dir_to_watch) / ".gitignore"
    if ignore_file is not None and ignore_file.exists() and ignore_file.is_file():
        ignore_file = str(ignore_file)
        with open(ignore_file, 'r') as f:
            gitignore = f.read()
        return PathSpec.from_lines(GitWildMatchPattern, gitignore.splitlines())

    return None


def maybe_requirements_txt_change(src_path: str, rpc_client: HotReloadMLRPCClient) -> Optional[MLRPCResponse]:
    if src_path.rstrip("~").endswith("requirements.txt"):
        requirements = get_requirements_from_file(Path(src_path))
        if requirements:
            response = rpc_client.reinstall_requirements(requirements)
            return response


def hot_reload_on_change(dir_to_watch, rpc_client: HotReloadMLRPCClient, frequency_seconds: int = 1,
                         logging_function: Callable = None,
                         error_logging_function: Callable = None,
                         success_logging_function: Callable = None):
    logging_function = logging_function or print
    error_logging_function = error_logging_function or print
    event_queue: queue.Queue[FileSystemEvent] = queue.Queue()

    logging_function("Attempting to start hot reload...")

    class FileChangeHandler(FileSystemEventHandler):
        def on_any_event(self, event: FileSystemEvent):
            event_queue.put(event)

    def consumer():
        logging_function(f"Starting file watcher for {dir_to_watch}...")

        def handle_response(_response: MLRPCResponse | List[MLRPCResponse],
                            event_type: Literal["hot-reload", "reinstall-pip-requirements"]):
            if response.status_code != 200:
                error_logging_function(
                    f"Event: {event_type} failed with status: {_response.status_code} - {_response.body}")
                return

            success_logging_function(f"Event: {event_type} status: {_response.status_code} - {_response.body}")

        while True:
            # Collect all changes made in the last 5 seconds
            any_changes = []
            start_time = time.time()
            while time.time() - start_time < frequency_seconds:
                try:
                    event = event_queue.get(timeout=1)
                    any_changes.append(event)
                except queue.Empty:
                    pass

            # If there are any changes, call full_sync
            if any_changes:
                ignore_specs = get_ignore_specs_for_hotreload(dir_to_watch)
                valid_changes = []
                for change in any_changes:
                    # if ignore specs is there and doesnt match its valid file
                    if ignore_specs is not None and not ignore_specs.match_file(change.src_path) \
                            and change.is_directory is False and change.src_path.endswith("~") is False:
                        valid_changes.append(change)
                    # if ignore specs is not there, all changes are valid
                    if ignore_specs is None:
                        valid_changes.append(change)
                logging_function(f"Found changes: {len(any_changes)} - Valid changes: {len(valid_changes)}")
                if not valid_changes:
                    continue

                for change in valid_changes:
                    responses = maybe_requirements_txt_change(change.src_path, rpc_client)
                    if responses is not None:
                        if isinstance(responses, list):
                            for response in responses:
                                handle_response(response, "reinstall-pip-requirements")
                        else:
                            handle_response(responses, "reinstall-pip-requirements")

                logging_function(f"Files changed firing hot reload for {dir_to_watch}...")
                responses = rpc_client.hot_reload(dir_to_watch)

                if responses is not None:
                    if isinstance(responses, list):
                        for response in responses:
                            handle_response(response, "hot-reload")
                    else:
                        handle_response(responses, "hot-reload")

    consumer_thread = threading.Thread(target=consumer, daemon=True)
    consumer_thread.start()

    event_handler = FileChangeHandler()
    observer = Observer()
    observer.schedule(event_handler, dir_to_watch, recursive=True)
    observer.start()

    return [consumer_thread, observer]


def extract_second_box(log_line):
    matches = re.findall(r'\[.*?\]', log_line)
    if len(matches) >= 2:
        return matches[1][1:-1]  # remove brackets
    else:
        return None


def extract_iso_timestamp(possible_ts: Optional[str]):
    if possible_ts is None:
        return None
    try:
        naive_dt = parse(possible_ts, ignoretz=True)
        utc_dt = timezone('UTC').localize(naive_dt).replace(tzinfo=None)
        return utc_dt
    except ValueError:
        return None


class LogMonitor:
    def __init__(self, ws: WorkspaceClient,
                 endpoint_name: str,
                 from_beginning: bool = False,
                 logging_function: Optional[callable] = None,
                 ):
        self.ws_client = ws
        self.endpoint_name = endpoint_name
        self.model_names = []
        for se in ws.serving_endpoints.get(endpoint_name).config.served_entities:
            self.model_names.append(se.name)
        now = datetime.utcnow().replace(tzinfo=None)
        self._last_log_ts = {mn: None for mn in self.model_names} if from_beginning else {mn: now for mn in
                                                                                          self.model_names}
        if from_beginning is False:
            logging_function(f"Looking for logs from: {now}")
        self._logging_function = logging_function or print

    def print_logs_if_havent_been_seen(self):
        logs = []
        for model_name in self.model_names:
            resp = self.ws_client.serving_endpoints.logs(self.endpoint_name, model_name)
            # read logs backwards and stop when we see a timestamp we've seen before
            latest_ts = None
            for idx, line in enumerate(reversed(resp.logs.splitlines())):
                ts = extract_iso_timestamp(extract_second_box(line))
                if ts is None:
                    continue
                if latest_ts is None:
                    latest_ts = ts
                if self._last_log_ts[model_name] is not None and ts <= self._last_log_ts[model_name]:
                    break
                logs.append(line)
            self._last_log_ts[model_name] = latest_ts

            for line in reversed(logs):
                self._logging_function(f"[{model_name.upper()}] " + line)
        return logs


def make_log_monitor_thread(ws: WorkspaceClient,
                            endpoint_name: str,
                            from_beginning: bool = False,
                            logging_function: Optional[callable] = None):
    def _monitor():
        monitor = LogMonitor(ws, endpoint_name, from_beginning, logging_function=logging_function)
        while True:
            try:
                monitor.print_logs_if_havent_been_seen()
            except Exception as e:
                logging_function(f"Error in log monitor: {e}")
            finally:
                time.sleep(10)

    return threading.Thread(target=_monitor, daemon=True)
